#include "osihfe.h"
#include "hfe.h"
#if defined (_WIN32)
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#define BCD2BIN( a ) ( ((( a >> 4) & 0x0F) *10) + ( a & 0x0F) )
#define BIN2BCD( a ) ( ((a / 10) << 4) + a%10 )
#ifndef __min
#define __min(X,Y) ((X) < (Y) ? (X) : (Y))
#endif

#define IsValidBCD(a) (((( a >> 4) & 0x0F) <= 9) && ((a & 0x0F) <= 9) )
#define BUFSIZE 8192


extern FILE* logfile;

const char szHeaderSig[] = "HXCPICFE";


/* OSI OS65D disk image info */
//
// 5.25" disk, 40 tracks, 8  sector/track,  256 bytes/sector, 11 bits/byte (8E1) =80K/disk
// OSI uses 8E1 to give a max unformatted capacity of 2272 bytes per track (see below)
// However other bit encodings  (8N1) could give up to 2500 bytes/track 
//
// The standard speed for mini-floppy drives (5") is 300 rpm. Thus one 
// rotation of the disk is 200 mS.
//
// Stated baud-rate is 125k or 125000 bits/sec and one serial byte is 11 bits 
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (125000 x 0.2) / 11  =  2272 bytes  or 8.8 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (125000 x 0.19) / 11  =  2159 bytes  or 8.4 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
//
// Data clocked out at 125000 bps, 11 bit word =11363 cps. 1Mhz clk = 88 cycles
// per character or 176 at 2Mhz	 (Floppy flux rate is 250000 fps)

//---------------------------------------------------------------------------

// 8" disk, 77 tracks, 12 sectors/track, 256 bytes/sector, 11 bits/byte (8E1/8N1) = 231K/disk
// OSI uses 8E1 to give a max unformatted capacity of 3772 bytes/track (see below)
// However other bit encodings (8N1) could give up to 3900 bytes/track
//
// The standard speed for 8" floppy dives is 360 rpm. Thus one rotation of 
// the disk is 166.6ms
//
// Stated baud rate is 250K or 250000 bits/sec and one serial byte is 11 bits
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (250000 x 0.166 ) / 11 = 	3772 or 14.7 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (250000 x 0.156) / 11  =  3545 bytes  or 13.8 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
// 
// Data clocked out at 250000 bps, 11 bit word =22727 cps. 
// floppy flux rate is 500000 fps
//---------------------------------------------------------------------------
// Track 0 format
// 10 millisecond delay after index hole
// 0,1 load address of the track in hi,lo form
// 2, page count of how much data is written on track 0, usually 8
// 3+ data
// 
// (SEE UPDATE BELOW for real timing values) 
//
// Track 1-n Format	 
// 10 ms delay after index hole
// 0,1 -a 2 byte start code $43, $57
// 2,  -BCD track number
// 3,  -track type code (always $58)
// (sector data follows)
//
// There can be any mixture of various length sectors. The total page count can not
// exceed 8 pages(12 on 8") (8*256) if more than one sector is on a track
// Each sector is written in the following format:
// -previous sector length (4 if none before) times 800 microseconds of delay
// -sector start code $76
// -sector number in binary
// -sector length (#pages) in binary
// -sector data
// -(usually has end of sector mark? $47, $53 )
//
// Directory Format
// 2 sectors (1 & 2) on track 12 (8 for 8") hold the directory information.
// Each entry requires 8 bytes. There are a total of 64 entries. The entries are
// formatted as follows:
// 0-5 ASCII 6 character filename
// 6 BCD first track of file
// 7 BCD Last track of file
//
// OS65D disk layout 5.25"
// Trk 0 - DOS pt1	$2200
// Trk 1 - DOS pt2	$2A00
// Trk 2-6 9 1/2 digit BASIC $0200-8pg, $0A00-8pg, $1200-8pg, $1A00-8pg, $2200-1pg
// Trk 7-9 Assembler Editor $0200-8pg, $0A00-8pg, $1200-5pg
// Trk 10-11 Extended Monitor $1700-8pg, $1F00-4pg
// Trk 12 sector 1,2 - Directory page 1,2  $2E79-1pg, $2E79-1pg
// Trk 12 sector 3 - BASIC overlays	 $20C4-1pg
// Trk 12 sector 4 - GET/PUT overlays $2E79-1pg
// Trk 13 COPIER/TRACK0 utility		 $0200-5pg
// Trk 14-39 user/utility programs	 COMPAR trk 39, $0200-5pg

// OS65D disk layout 8"
// Trk 0 - DOS pt1-8pgs	$2200
// Trk 1 - DOS pt2-5pgs	$2A00
// Trk 1 sector 2, 5pgs Copier 
// Trk 2-4 9 1/2 digit BASIC $0200-11pg, $0D00-11pg, $1800-11pg
// Trk 5-7 Assembler Editor $200 /Extended Monitor@1700 $0200-12pg, $0E00-12pg, $1A00-7pg
// Trk 7 sector 2 - unused
// Trk 7 sector 3 - unused
// Trk 7 sector 4 - ??
// Trk 7 sector 5 - unused
// Trk 7 sector 6 -	unused

// Trk 8  sector 1,2 - Directory page 1,2  $2E79-1pg, $2E79-1pg
// Trk 8 sector 3 - BASIC overlays	 $20C4-1pg
// Trk 8 sector 4 - GET/PUT overlays $2E79-1pg
// Trk 8 sector 5-8pgs
// Trk 9-11 11pgs/trk BEXEC*
// Trk 12 Colors
// Trk 13 COPIER/TRACK0 utility		 $0200-5pg
// Trk 14-39 user/utility programs	 COMPAR trk 39, $0200-5pg


//------------------------------------------------------------------------------

// OS65U (for Business systems)
// OS65U sector size is 3584, one sector per track on 8" floppy 
//
// Track 0 format
// 10 millisecond delay after index hole 8E1
// 0,1 load address of the track in hi,lo form
// 2, page count of how much data is written on track 0, usually 13
// 3+ data
//
// Track 1-76 format
// wait for index hole, set 8E1
// $80, tkHi, tkLo, Reset ACIA 8N1 +few ms delay
// <sometimes noise with hi bit set> $0F or $1F marker?, $E00 (3584) bytes data + 2bytes checksum (data & header)
//
// File entries 1st 16 bytes contain a copy of the diretory entry
//

// -------------------------------------------------------
// HEXDOS operating system for OSI C1 series
// Track 0 follows normal OSI boot track rules ecxept there are 9 pages of data
// loading at $0300
//
// Track 01-39
// $53, trk# (in binary) <2048 bytes of data>
//directory track is track 1
//directory consists of basic compatible listing loaded at $0B00
//@$0B00:                                    
//00,   0C ,  0B  ,  02  ,  00  ,HEXDOS, 00, 12, 0B, 09, 00 , A5, 00, 00, 00, 00, 00, 00
//00, ptrLo, ptrHi, trkLo, trkHi, <name>,00, 
// ptrLo, ptrHi points to next entry unless 0 which indicates end of list
// the last entry is a *(A5?) indicating the 1st available track
 
// DOS/65
// Track 0-n Format
// 10 ms delay after index hole
// 0,1 -a 2 byte start code $43, $57
// 2,  -BCD track number
// 3,  -track type code (always $58)
// -sector start code $76
// -sector number in binary	 (always 1)
// -sector length (#pages) in binary (always 8 (or 13 on 8"))
// -sector data	(up to 8*256, 13*256)


/*
  (from T.R. Berger in Compute's Gazette #20 Jan 1982 p136)
*** Actual OS65D disk timing ***

 The data on track 1+ commences 1ms (1000us) past index hole, about 23 bytes in time at 44us/byte
 2 bytes are written to indicate start of track. $43, $57. When the track is read, reading
 does not commence until $43 $57 have been found. Next the track number in BCD is written,
 then a stop byte $58 is written (it is never checked on read). This constitutes the track
 header. On initialization, the track is erased, then the track header is written. It is not
 rerecorded on any other read or write.
 There is a lull after track header of 6.6ms (about 149 bytes) This time differs greatly from
 the time given in the OS65D Guide!
 During a sector seek operation, a "previous sector length" number is saved "p". This value
 is set to 4 for a sector 1 seek. Then a subroutine waits p * 800us. The OS65D Guide says
 that between sector N and Sector N+1 there is a p * 800ms gap. This is not quite correct.
 After the end of sector OS65D waits p * 800us, the write function is switched on, a further
 185us is allowed to pass then the erase function is switched on. we now wait an additional
 p * 800us before starting to write data. In other words the time from the last byte of sector
 N to the first byte of sector N+1 is about p * 1600us + 185us. For sector 1 p is taken to be 4
 In all other cases "p" is the length in pages (256bytes) of the previous sector.
 This applies to 1MHz systems. On boot OS65D measures the timing on the serial port to measure
 clock speed, then a timing constant in the 1ms subroutine is set. However the timing calculation
 does not affect the 100us routine used in sector spacing. So the 100us routine is really a
 100us/T where T=the MHz of the machine. This calculation only accounts for the wait loops in
 intersector timing. There is quite a lot of linear code. A crude estimate would be to add an
 additional 30us after each sector. In other words the sector spacing is (p * 1600+215)/T us.
 where p is the number of pages in the previous sector and T is the system speed in MHz.
 All of this says there is some kind of empty space between the track header and the 1st sector.
 Each sector is completely rewritten each time it is addressed in a write operation.
 A sector is written as follows: A sector start code $76 is written, next the sector number and
 the sector length in pages "p".  The sector is verified with a read operation against the value
 in memory. Next is the data, p*256 bytes long. followed by a sector gap already covered.
 When writing sector N, the computer must read the preceeding sectors. While searchig for a sector
 it also watches for index hole again to keep from hanging on a sector read. After the start of the
 preceeding sector, the computer no longer searches for an index hole as there is not enough time
 to watch for end and prepare for the next sector.  If the index hole passes during a write operation
 the header may be overwritten rendering the track unreadable.
 So on a 1MHz machine with 8" floppies spinning at 360 RPM, one revolution takes 166,667us
 the disk data is clocked at 250KHZ each bit takes 4us. Since OSI bytes uses 11bits, 44us are 
 required per byte. If we could pack a track we could fit 3787 bytes. But it is not packed
 accounting for formatting time we use 1000ms from the index hole, The header is 4bytes long
 using 176us more. The gap to sector 1 is 4*1600+215 or 6615us, so 7791us are used between
 the index and start of sector 1. Each sector contains an integral number of pages, thus all
 sectors contain as an aggregate p pages. Each byte take 44us and there are 256 bytes per page
 thus all these pages take p*11264us. Each sector takes 5 extra bytes thus for n sectors we have 
 n * 220us. Next we must account for all the wait time after each of the n sectors.
 The wait time for one sector to the next is q * 1600+215us where q is the number of pages in
 the preceeding sector. Since we assume there are n sectors on a track there are only n-1
 spaces between sectors. If the last sector has r pages, then the preceding n-1 contain
 p-r pages altogether. Thus the total intersector wait time is 1600*(p-4)+215*(n-1)us. Finally
 we must account for the time after the last sector is written before the erase functions are 
 switched off. Write and erase continue for r*600us after the last byte is written. The write is
 switched off and erase continues for 600us more before it is switched off. This total trailing
 time is 525+600*r us.
 For 8" floppies with a 1MHz clock the total recording time for n sectors is:
 Tus=8101+12864*p -1000*r +435*n us, where p is the total numer of pages, and r is the number
 of pages in the last sector
 For 5.25" floppys clock data at 125KHz and spin at 300RPM

 8 INCH FLOPPY TIMING
 # Sectors   Total    Pages         Time      DT
             Pages 	 Last Sector
	  1		  13	    13		   162768	 3900
	  2		  13		10		   166203	  464
	  3		  13		10		   166638	   29
	  4		  12		1		   163209	 3458
	  5		  12		1		   163144	 3028
	  6		  12		1		   164079	 2588
	  7		  12		1		   164514	 2153
	  8		  12		1		   164949	 1718
	  9		  12		1		   165384	 1283
	 10		  12		1		   165819	  848
	 11		  12		1		   166254	  413
	 12		  12		1		   166689	  -22
	 t(us.) = 8101 + 12864*p - 1000*r + 435*n
	 p = number of pages, r = number of pages in last sector, n=number of sectors
	 166667us = time on one track
	 DT = time left on track

 5.25 INCH MINI FLOPPY TIMING
 # Sectors   Total    Pages         Time      DT
             Pages 	 Last Sector
								   
	1		  8			8		   193986	  6014
	2		  8			3		   199641	   359
	3		  8			4		   199296	   704
	4		  8			4		   199951		49
	5		  7			1		   179478	 20522
	6		  7			1		   180133	 19867
	7		  7			1		   180788	 19212
	8		  8			1		   205571	 -5571
	t(us.) = 8307 + 24128*p -1000*r + 435*n
	200000 us = time on one track
*/

//index to bitmask 
const unsigned char mask[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };


//FM encoding state
struct FMOutData
{
	//eSERFORMAT eSformat; //serial word format
	unsigned char data;	//current output character
	unsigned int nPos;	//bit index

};

//OSI OS65D sectors can be any combination of up to 8 or 13 sectors (5.25/8") totalling 8/13 256 byte pages
// So you can have 8 1-page sectors. 2 4-page sectors , or a 1,2,1,1,3 page sector track on 5.25"
struct OSISector
{
	int nBlockLen[20]; //length of sector
	const unsigned char * pData[20];  //pointer to sector data
};



//find first occurrence of pFind in pSearch of specified lengths 
unsigned char * bStrStr( unsigned char * pSearch, unsigned char *pFind, int nSearchLen, int FindLen)
{
	if (FindLen > nSearchLen || !FindLen || !nSearchLen || !pSearch || !pFind)
		return NULL;
	int pos, matchpos;
	for (pos = 0; pos + FindLen < nSearchLen; pos++)
	{
		if ( (pSearch[pos] ) == pFind[0])
		{
			matchpos = 1;
			while (matchpos < FindLen)
			{
				if ( (pSearch[pos+matchpos]) != pFind[matchpos])
				{
					break;
				}
				matchpos++;
			}
			if (matchpos == FindLen)
			{
				return pSearch + pos;	
			}
		}
	}
	return NULL; 
}
//find last occurance of pFind in pSearch
const unsigned char * bStrRStr(const unsigned char * pSearch, const unsigned char *pFind, int nSearchLen, int FindLen)
{
	if (FindLen > nSearchLen || !FindLen || !nSearchLen || !pSearch || !pFind)
		return NULL;

	int pos, matchpos;
	for (pos = nSearchLen - FindLen; pos >=0; pos--)
	{	
		if ( (pSearch[pos] ) == pFind[0])
		{
			matchpos = 1;
			while (matchpos < FindLen)
			{
				if ( (pSearch[pos+matchpos]) != pFind[matchpos])
				{
					break;
				}
				matchpos++;
			}
			if (matchpos == FindLen)
			{
				return pSearch + pos;	
			}
		}
	}
	return NULL; 
}
// int handleData(FMstate& fmstate, int pos, bool bVerbose)
//
//return decoded character if >= 0
//return contine otherwise
//
//pos indicates position in input data for data logging
//fmstate is the current serial data state
//bVerbose enables extended data logging 
//
int handleData(FMstate& fmstate, int pos, bool bVerbose)
{
	int nRet = -1;
	int nParity = 0;
	int i;
	
	//start bit is 0, stop bit is 1, but bits are INVERTED due to ACIA

	switch(fmstate.eSformat)
	{
		default:
		case s8E1: //start bit = bit 0, data 1..8, parity 9, stop 10
				if (fmstate.sWordIndex >= 11)
				{
					//check for correct framing (inverted)
					if (fmstate.bSerWord[0] == 1 && fmstate.bSerWord[10] == 0)
					{
						nRet = 0;

						//decode serial byte
						for (int i = 8; i >= 1; i--)
						{
							
							nRet = (nRet << 1);
							if (!fmstate.bSerWord[i])
							{
								nRet |= 1;
								nParity++;
							}
						}
						// even parity - sum of data bits + parity is even
						if (!fmstate.bSerWord[9])
							nParity++;

						if (nParity & 1)
						{
							fmstate.error |= ERR_PARITY;
							//Parity Error
							if (bVerbose)
								LogText( _T("PE @ %05lX\r\n"),  pos);
								
						}
						else
						{
							fmstate.error &= ~ERR_PARITY;
						}

						fmstate.sWordIndex = 0;
					}
					else
					{
						//keep shifting lsb out until framing works
						for ( i = 1; i <= 10; i++)
						{
							fmstate.bSerWord[i-1] = fmstate.bSerWord[i];
						}
						fmstate.bSerWord[10] = 0; 	
						fmstate.sWordIndex = 10;


					}
				}
			break;

		case s8N1:	//start bit = bit 0, data 1-8, stop bit 9
				if (fmstate.sWordIndex >= 10)
				{
					if (fmstate.bSerWord[0] == 1 && fmstate.bSerWord[9] == 0)
					{
						nRet = 0;
						//decode serial byte
						for (int i = 8; i >= 1; i--)
						{
							nRet = (nRet << 1);
							if (!fmstate.bSerWord[i])
								nRet |= 1;
						}
						fmstate.sWordIndex = 0;
						fmstate.error &= ~ERR_PARITY;

					}
					else
					{
						//shift lsb out until framing works
						for ( i = 1; i <= 9; i++)
						{
							fmstate.bSerWord[i-1] = fmstate.bSerWord[i];
						}
						fmstate.bSerWord[9] = 0; 	
						fmstate.sWordIndex = 9;
					}	
				}



			break;
	}

	return nRet;	
	
}

void FMdecodeInitialize(FMstate& fmstate)
{
	memset( &fmstate, 0, sizeof(FMstate));
	fmstate.eSformat = s8E1;
	
}


// - decoded data may be spread across multiple calls to FMdecodestream, i.e. there may be incomplete decodes
//   when inBuf is exhausted.
// - not all input data may be consumed in one call to FMdecodestream if outBuf is too small
//
// - looks like stream of 001000100010001000100010... is clock bit with no (0) data and
//              stream of 101010101010101010101010... is clock bit with a (1) data, 2 bits HFE data to 1bit data
//                   
int FMdecodestream(const unsigned char *inBuf, int nInLen, int bitwidth, unsigned char* outBuf, unsigned char *errBuf, int nOutLen, FMstate& fmstate, int* pnRead, int *pnWrote, int pos, bool bVerbose)
{
//assume 1st "1" bit is clock, 2nd bit is data. Assume lsb of inBuf	is 1st bit written to FDC
//HFE FM data is 2bits for each clock or data bit
//assume OSI data is stored stored bit 0 first (lsb) in serial word format (10 or 11 bits)
//the start bit of the serial word is on disk is a '1' because serial data is INVERTED
//data bits can be spread between blocks, a clock could be at the end of one, data at the next
//but generally all data is nibble aligned

	int nRead = 0;
	int nWrote = 0;
	bool bcont = true;

	//read bits from inBuf characters, until a '1' is encountered
	//that is clock bit
	//read next bit into recv buf, until 1 serial word is recovered
	//check for serial word framing, parity etc. errors
	//write decoded serial word to output buffer
	while( nRead < nInLen && bcont)
	{
		int nChar = 0;

		if (fmstate.nPos == 0)
		{
			fmstate.data = inBuf[nRead++]; //get next byte to process
			pos++;
		}
		//while processing input byte
		while (fmstate.nPos < 8)
		{
			bool b = (fmstate.data & mask[fmstate.nPos++]) > 0;

			if (bitwidth > 1 && fmstate.nSampleCount++ < 0)	 //we only look at every other bit from HFE data after syncing
			{			
				continue;
			}
			
			if (fmstate.bClock)
			{
				//FM clock has been detected, now read data
				//if bitwidth > 1, if next 3 bits are 0, then data bit is 0 else if at least one of next 3 bits is 1, data is 1
				if (!b)
				{
					if (bitwidth > 1 && fmstate.nSampleCount < 1) //just skipping 0's
					{
						continue;
					}
				}
				//reading data
				fmstate.bSerWord[fmstate.sWordIndex++] = b;
				fmstate.bClock = false;		//got data, now wait for FM clock
				fmstate.nSampleCount = -1;	//eat next bit
				if (!b)
					fmstate.nZeroRun++;
				else 
				{
					if (fmstate.nZeroRun > 20)
					{
						if (bVerbose)
							LogText( _T(" Zeros %02d \r\n"),fmstate.nZeroRun);
						fmstate.bPastIndex = true;
					}
					fmstate.nZeroRun = 0;
				}
					

				nChar = handleData(fmstate, pos, bVerbose);  //process bit from data stream
				if (nChar >= 0 )
				{
					//are we done with a byte out?
					

					if (fmstate.bPastIndex)
					{
						
						if (bVerbose)
						{
							LogData((unsigned char) nChar);

						}						
						*outBuf++ = (unsigned char) nChar;
						if (errBuf)
						{
							*errBuf++ = fmstate.error;
						}
						fmstate.error = 0;

						if (++nWrote >= nOutLen)
						{
							//outbuf full
							bcont = false;
							break;
						}
					}
					else
					{
						if (bVerbose)
							 LogText(_T("(%02X)"), (unsigned char)  nChar);	
					}
				}
			}
			else
			{
				//we're expecting clock
				if (b)
				{					
					//have clock
					fmstate.bClock = true;
					//when expecting clock, on 1st transition, eat next bit, whatever it is
					fmstate.nSampleCount = -1;
					if (fmstate.nulRun > 100)
					{
						if (bVerbose)
							LogText( _T(" Null %02d \r\n"),fmstate.nulRun);
						fmstate.error = 0;
						
					}
					fmstate.nulRun = 0;
				}
				else
				{
					fmstate.nulRun++;
				}
			}			
		}
		fmstate.nPos = 0;

	} //while there is input

	if (pnRead)
	{
		*pnRead = nRead;
	}
	if (pnWrote)
	{
		*pnWrote = nWrote; 
	}
	return 1;

}

//write 1 data bit,  FM clock & data = 4 flux bits (lsb 1st)
int WriteNibble(bool bWriteBit, FMOutData& fmd, unsigned char* lpOut)
{
	//one FM clock per data bit
	unsigned char c = 0;
	int nWrote = 0;

	if (bWriteBit)
		c = (unsigned char)( (fmd.nPos > 0) ? 0xA0 : 0x0A); //1010 
	else				   //					  ~~ clock
		c = (unsigned char)((fmd.nPos > 0) ? 0x20 : 0x02); //0010
	if (fmd.nPos > 0)					 //	~~	 data
	{	
		//finish writing last bit
		*lpOut++ = (unsigned char) (fmd.data | c);
		fmd.nPos = 0;
		fmd.data = 0;

		nWrote++;
	}
	else
	{
		fmd.nPos = 4;
		fmd.data = c;
	}
	return nWrote;
}

//write 1 data bit,  FM clock & data = 2 flux bits (lsb 1st)
int WriteBit(bool bWriteBit, FMOutData& fmd, unsigned char* lpOut)
{
	//one FM clock per data bit
	unsigned char c = 0;
	int nWrote = 0;

	if (bWriteBit)
	{
		switch (fmd.nPos)
		{
			default:
			case 0:
				c = (unsigned char) 0x03;  //clock + data lsb first
				break;
			case 1:
				c = (unsigned char) 0x0C;
				break;
			case 2:
				c = (unsigned char) 0x30;
				break;
			case 3:
				c = (unsigned char)0xC0;
				break;
		}
		
	}
	else
	{
		switch (fmd.nPos)
		{
			default:
			case 0:
				c = (unsigned char) 0x01;  //clock 1st, no data
				break;
			case 1:
				c = (unsigned char) 0x04;
				break;
			case 2:
				c = (unsigned char) 0x10;
				break;
			case 3:
				c = (unsigned char)0x40;
				break;
		}
	}				//					  
	if (fmd.nPos >= 3)
	{	
		//finish writing last bit
		*lpOut++ = (unsigned char) (fmd.data | c);
		fmd.nPos = 0;
		fmd.data = 0;

		nWrote++;
	}
	else
	{
		fmd.nPos++;
		fmd.data |= c;
	}
	return nWrote;
}

// Write a bitlen number of FM encoded 0's, return # character written;
// one clock pulse before each 0, 4 bits out for each 0 bit
int FMWriteZeros( int bitlen, FMOutData fmd, int bitwidth, unsigned char* lpOut, int nOutLen)
{
    //one FMclock + 0 is 4 bits out	 0100
	unsigned char c = 0;
	int nWrote = 0;

	if (bitwidth == 1)
	{
		//finish writing last bit
		while (fmd.nPos > 0 && bitlen > 0)
		{
			nWrote += WriteBit(0, fmd, lpOut);
			bitlen--;
		}

		while (bitlen > 4 && nWrote+1 < nOutLen)
		{
			//write a bunch of quads
			*lpOut++ = 0x55;
			bitlen -= 4;
			nWrote++;
		}
		while (bitlen > 0)
		{
			nWrote += WriteBit(0, fmd, lpOut);
			bitlen--;
		}
	}
	else
	{
	
		if (fmd.nPos > 0)
		{								     //00100010
			//finish writing last bit
			c = (unsigned char) ((fmd.data & 0x0F) | 0x20);	 //0010xxxx
			fmd.nPos = 0;
			fmd.data = 0;
			*lpOut++ = c;
			bitlen--;
			nWrote++;
		}

		while (bitlen > 2 && nWrote < nOutLen)
		{
			//write a bunch of pairs
			*lpOut++ = 0x22;
			bitlen -= 2;
			nWrote++;
		}
		if (bitlen > 0)
		{
			//last nibble is stored in the buffer
			fmd.data = 0x02;
			fmd.nPos = 4;

		}
	}

	return nWrote;

}

// Write a byte in FM + 8E1 format (11 bits data, 44 output bits, 5.5 bytes) 
int FMWrite8E1(unsigned char data, FMOutData& fmd, int bitwidth, unsigned char* lpOut, int OutLen)
{
	int nWrote = 0;
	int nParity = 0;
	(void)OutLen;   //Do nothing

	//(data is inverted upon write)
	//write start bit
	if (bitwidth == 1)
		nWrote = WriteBit( 1, fmd, lpOut);
	else
		nWrote = WriteNibble(1, fmd, lpOut);

	//write data bits
	for (int i = 0; i < 8; i++)
	{
		bool bSet = (data & mask[i]) > 0;

		if (bitwidth == 1)
			nWrote += WriteBit( !bSet, fmd, lpOut+nWrote);
		else
			nWrote += WriteNibble( !bSet, fmd, lpOut+nWrote);
		if (bSet)
			nParity++;
	}
	//write parity
	if (bitwidth == 1)
		nWrote += WriteBit( !(nParity & 1), fmd, lpOut+nWrote);
	else
		nWrote += WriteNibble( !(nParity & 1), fmd, lpOut+nWrote);

    //stop bit
	if (bitwidth == 1)
		nWrote += WriteBit( 0, fmd, lpOut+nWrote);
	else
		nWrote += WriteNibble( 0, fmd, lpOut+nWrote);
	
	return nWrote;

}

// Write a byte in FM + 8N1 format (10 bits)
int FMWrite8N1(unsigned char data, FMOutData& fmd, int bitwidth, unsigned char* lpOut, int OutLen)
{

	int nWrote = 0;
	(void)OutLen;   //Do nothing


	//(data is inverted upon write)
	//write start bit
	if (bitwidth == 1)
		nWrote = WriteBit(1, fmd, lpOut);
	else
		nWrote = WriteNibble(1, fmd, lpOut);

	//write data bits
	for (int i = 0; i < 8; i++)
	{
		bool bSet = (data & mask[i]) > 0;
	    if (bitwidth == 1)
			nWrote += WriteBit( !bSet, fmd, lpOut+nWrote);
		else
			nWrote += WriteNibble( !bSet, fmd, lpOut+nWrote);
	}

    //write 1 stop bit
	if (bitwidth == 1)
		nWrote += WriteBit( 0, fmd, lpOut+nWrote);
	else
		nWrote += WriteNibble( 0, fmd, lpOut+nWrote);
	
	return nWrote;

}

//locate start of track data for OS65D, Hexdos, os65U
//there may be garbage bytes before expected  data
int OSISeekTrackStart( const unsigned char* pTrack, int nLen, DISKTYPE dt /*= DISK_UNKNOWN*/, int iTrack /*=-1*/)
{
	int i = 0;
	int guess = -1;
	
	if (!pTrack)
		return -1;

	if (nLen > 32)  //limit search to reasonable length
		nLen = 32;

	while (i+4 <= nLen)
	{
		if (pTrack[i] == 0x43 && pTrack[i+1] == 0x57 && (iTrack >= 0 ? pTrack[i+2] == BIN2BCD(iTrack) : pTrack[i+2] > 0) &&  dt != OS65U8) //OS65D
		{
			unsigned char c = pTrack[i+2];
			int nTrack = BCD2BIN(c); //BCD to decimal
			
			if ( ((iTrack > 0 && iTrack == nTrack) || (iTrack < 0 && nTrack < 80)) && (pTrack[i+3] == 0x58))
			{
				return i;

			}
			else
			{
				//not valid OS65D track# (5.25 max = 39, 8" = 77)
				guess = i;
				
			}	
		}
		else if (pTrack[i] == 0x57 && ( iTrack >= 0 ? pTrack[i+1] == iTrack : pTrack[i+1] < 40) && dt != OS65U8 && (i == 0 || pTrack[i-1] != 0x43) )
		{	//this detection is too ambiguous, one wayward byte and 
			//hexdos
			return i;
		}
		else if (pTrack[i] == 0x80 && pTrack[i+1] == 0)
		{
			int nTrack = pTrack[i+2];
			if (iTrack >= 0 && iTrack == nTrack)
			{
				if (dt == OS65U8 || dt == DISK_UNKNOWN)
					return i;
				
			} else if (iTrack < 0 && nTrack < 80)
			{
				guess = i;
				if (dt == OS65U8) //if we're sure we are looking for this, end!
					break; 
			}
		}
		else if (pTrack[i] == 0x43 && pTrack[i+1] == 0x57 && pTrack[i+3] == 0x00 )
		{
			//could be CPM 2.2 variant  0x43,0x57,0xTRK, 0x00,<data>
			int nTrack = pTrack[i+2];
			if (iTrack >= 0 && nTrack == iTrack)
			{
				return i;
			}
			else if (nTrack < 80)
			{
				guess = i;

			}

		}
		
		i++;
	}

	return guess;
}

//return array of OSI OS65D/OS65U/HexDos/Dos65 sector lengths & pointers to sector data
int OSICountSectors(const unsigned char * pTrack, int nTrkLen, int nTrack, OSISector& sect)
{
	memset(&sect, 0, sizeof(OSISector));
	int nCount = 0;

	if (!pTrack)
		return 0;

	const unsigned char *p = pTrack;
	int i = 0, nStart = 0;


	i = OSISeekTrackStart(p, nTrkLen, DISK_UNKNOWN , nTrack);

	if (i >= 0 && pTrack[i] == 0x43)
	{
		
		if (pTrack[i+1] == 0x57 && pTrack[i+3] == 0x00 &&  (nTrack >= 0 ? pTrack[i+2] == nTrack : pTrack[i+2] < 80 ) )
		{
			//Probable CPM 2.2 variant
			sect.nBlockLen[0] = 13;
			sect.pData[0] = (pTrack + i + 4);
			return 1;


		}
		i=i+4; //skip start of track
		nStart = i;
		// $76, sect#, sect-len 
		while (i+5 < nTrkLen)
		{
			if ((pTrack[i] == 0x76 && nCount == 0 && i-nStart < 20) || //start of sector marker	near start of sector
			    (pTrack[i] == 0x76 && nCount > 0 && pTrack[i+1] == nCount+1) )
			{
				int nSect = pTrack[i+1];
				int nSectLen = pTrack[i+2];

				if (nSect != nCount+1 || nSectLen > (nTrkLen/256) )
				{ 
					//bad sect 1
					break;
				}


				if (nSect < 14 && nSect > 0 && nSectLen <= 14 && nSectLen > 0 && nSect > nCount )  //don't allow multiple sectors of same#
				{
					//looks legit
					sect.nBlockLen[nCount] = nSectLen;
					sect.pData[nCount] = (pTrack + i + 3);

					i = i + nSectLen*256 + 3;
					
					//now expect $47,$53 

					if (i+1 > nTrkLen)
					{
						break;	//don't include sector that runs off end of track
					}
					nCount++;

					if ( pTrack[i] == 0x47 && pTrack[i+1] == 0x53)
					{
						//end of sector mark sometimes present
						i++;

						//now there could be more sectors or 0's or whatever
					}

				}
				
			}
			i++;
		}
		
		
	}
	else if (pTrack[i] == 0x80 && pTrack[i+1] == 0x00)
	{
		//OS65U
		sect.nBlockLen[0] = 0xE03;
	    
		i+=3;
		if (pTrack[i] & 0x80) //skip possible noise
			i++;
		sect.pData[0] = (pTrack + i);
		nCount = 1;
	}
	else if (pTrack[i] == 0x57 && pTrack[i+1] < 40 && (i == 0 || pTrack[i-1] != 0x43) )
	{
		//hexdos
		sect.nBlockLen[0] = 2048;	
		i+=2;
		sect.pData[0] = (pTrack + i);
		nCount = 1;
	}
	
	return nCount; 
}

// pTrack points to OSI track data of TRKSIZ8 or TRKSIZ5 bytes, or NULL to create blank track
// DISKTYPE specifies disk type & thus size of pTrack
// nTrack specifies OSI track #, track 0 is handled differently

int FMEncodeOSITrack( unsigned char * pTrack, DISKTYPE dt, int nTrack,  int bitwidth, unsigned char * lpOut, int nOutLen, bool bWrite8As5)
{
	//Convert from WinOSI disk image to HFE disk image
	//must replace timing delays lost in OSI dump process
	//
	// timing after index hole, timing between encoding changes (8E1/8N1 OD65U)
	// timing before sectors OS65D
	// must extract clean data from WinOSI image

	if (bitwidth != 1)
		bitwidth = 2;



	if (!lpOut || nOutLen < 1)
		return -1;

	int nInLen = 0;
	int nPos = 0;
	int fpm = 250;		//flux per millisecond 
	int nDisk8Index = DISK8INDEX;
	FMOutData fmd;
	unsigned char BlankBuf[TRKSIZ8];

	//fmd.eSformat = s8E1;
	fmd.data = 0;
	fmd.nPos = 0;
	int nDiskTargetLen = 0;

	//track buffer needs up to 
	// 500000 bps * 166.6ms	= 83333 flux per track.	(data clock rate 250K)
	//  @ 2 flux per bitcell (8bits of memory) = 41666 bytes for 8" track
	// 250000 bps * 200ms = 50000 flux per track.
	// @ 2 flux per bitcell = 25000 bytes for 5.25" track
	// or so we think

	//index pulse width on 8" drive is ~1.7ms so about 850 zeros
	//from OSI formatted disk on HxC 5.25" has ~1478 zeros  (expected less??)
	//from OSI formatted disk on HxC 8" has ~757 zeros
	//fron Kryoflux image, 8" drive has about 250 zeros	(TOO LOW)
	OSISector sect;
	memset(&sect, 0, sizeof(OSISector));
	
	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
		case EMPTY5:
		case EMPTY5Z:
			nInLen = TRKSIZ5;
			nDiskTargetLen = DISK5TARGETLEN;
			break;

		case OS65D8:
		case DOS658:
		case OS65U8:
		case EMPTY8:
		case EMPTY8Z:
			nInLen = TRKSIZ8;
			nDiskTargetLen = DISK8TARGETLEN;
			fpm =500;
			break;
		default:
			break;
	}
	if (bWrite8As5)
	{
		nDisk8Index = DISK5INDEX;
	}
	if ( bitwidth == 1)
		nDiskTargetLen /=2;

	if (dt == EMPTY5 || dt == EMPTY8 || dt == EMPTY5Z || dt == EMPTY8Z)
	{
		nPos =  FMWriteZeros( nDiskTargetLen*2 , fmd, bitwidth, lpOut, nOutLen);
		if (dt == EMPTY5Z || dt == EMPTY8Z)
			memset(lpOut, 0, nPos);
		
		return nPos;
	} 
	int nChk = 	(*(pTrack+nInLen-3) << 8) | *(pTrack+nInLen-4);					
	int curTrackLen =  (*(pTrack+nInLen-1) << 8) | *(pTrack+nInLen-2);
	if ((nChk ^ 0xFFFF) != curTrackLen || curTrackLen == 0)
		curTrackLen = -1;

	if (nTrack == 0)
	{
		//special handling for track 0
		if (dt == OS65D5 || dt == HEXDOS)
		{
			//write leader
			nPos = FMWriteZeros( DISK5INDEX, fmd, bitwidth, lpOut, nOutLen);
			int nLen = 8*256 + 3;
			if (dt == HEXDOS)
				nLen += 256;	//hexdos uses extra space after sector 8 on boot for autorun
			if (curTrackLen > 0)
				nLen = curTrackLen;

			//write data
			for (int i = 0; i <	nLen && i < nInLen; i++)
			{
				if (pTrack)
					nPos += FMWrite8E1(pTrack[i], fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				else
					nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

			}
			//write track filler
			nPos +=  FMWriteZeros( (nDiskTargetLen - nPos)*2 , fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
		}
		else //8"
		{
			//write leader
			nPos = FMWriteZeros( nDisk8Index, fmd, bitwidth, lpOut, nOutLen);
			int nLen = 8*256;
			
			if ( pTrack)
			{
				if (pTrack[2] > 8 && pTrack[2] < 14)
					nLen = pTrack[2] * 256;
				nLen += 3;
			}
			else
				nLen = 3 + 256*13;

			for (int i = 0; i <	nLen && i < nInLen; i++)
			{
				if (pTrack)
					nPos += FMWrite8E1(pTrack[i], fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				else
					nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			}
			//write track filler
			nPos +=  FMWriteZeros( (nDiskTargetLen - nPos) *2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);


		}

	}
	else
	{
		if (dt == HEXDOS)
		{
			int nSectors = 0;
			if (pTrack)
				nSectors = OSICountSectors( pTrack, TRKSIZ5 , nTrack, sect);
			else
			{				
				//we're going to make 1 blank track size sector
				//<index><delay>$57,trk#(binary),$896 bytes data, ends with $FF,$FF
 				nSectors = 1;
				memset(&BlankBuf[0], 0, sizeof(BlankBuf));
				sect.pData[0] = &BlankBuf[0];
				sect.nBlockLen[0] = 898;
			    BlankBuf[896] = 0xFF; 
			    BlankBuf[897] = 0xFF; 
			}
			
			//write leader & track id
			nPos = FMWriteZeros( DISK5INDEX , fmd, bitwidth, lpOut, nOutLen);
			nPos += FMWrite8E1(0x57, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1((unsigned char)nTrack, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			 
			if (nSectors > 0) 
			{
				const unsigned char * p = sect.pData[0];

				//write sector data
				for (int i = 0; i < sect.nBlockLen[0]; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				
				}

				//write trailing bits or padding
				nPos += FMWriteZeros( (nDiskTargetLen - nPos) * 2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

			 }
			 else
			 {		//no sectors, unknown or blank formatting

				const unsigned char * p = pTrack ? pTrack+2 : NULL; //skip track marker

				//write unknown track data, no special formatting
				for (int i = 0; i <  2048; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

				}
				//write trailing bits or padding
				nPos += FMWriteZeros( (( nDiskTargetLen ) - nPos)*2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);


			}

		}
		else if (dt == OS65D5 || dt == OS65D8 || dt == DOS655 || dt == DOS658)
		{
		    //OS65D
			int nSectors = 0;
			if (pTrack)
				nSectors = OSICountSectors( pTrack, dt == OS65D5 ? TRKSIZ5 : TRKSIZ8, nTrack, sect);
			else
			{
				//we're going to make 1 blank track size sector 
				nSectors = 1;
				memset(&BlankBuf[0], 0, sizeof(BlankBuf));
				sect.pData[0] = &BlankBuf[0];
				sect.nBlockLen[0] = (dt == OS65D5 ? SECTORS5 : SECTORS8);
			    BlankBuf[0] = 0x76; 
			    BlankBuf[1] = 0x01; 
			    BlankBuf[2] = (unsigned char) (dt == OS65D5 ? SECTORS5 : SECTORS8);
				int end = BlankBuf[2] * 256 + 3;
				BlankBuf[end] = 0x47;
				BlankBuf[end+1] = 0x53; 

			}

			int nCount = 0;
			int lastlen = 1;
			int nBCDTrk =  BIN2BCD(nTrack); 
			
			//write leader & track id
			nPos = FMWriteZeros( dt == OS65D5 ? DISK5INDEX : nDisk8Index, fmd, bitwidth, lpOut, nOutLen);
			nPos += FMWrite8E1(0x43, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1(0x57, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1((unsigned char)nBCDTrk, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1(0x58, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
			 
			 
			if (nSectors > 0)
			{
				while (nCount < nSectors && nOutLen-nPos > 0)
				{
					//fpm = flux per millisecond / 10 = flux per microsecond
					//now delay = previous sector length (or 4 if first) times .8ms delay
					
					//UPDATED: at 1MHz CPU clock delay after index hole is 1000us
					//delay between track header and 1st sector is 6600us !
					//delay between sectors is previous sector length in pages * 1600us + 185us
					
					int nDelay = (int)((fpm * lastlen));  //now a little longer	was (fpm * lastlen)/1.25

					nPos += FMWriteZeros( nDelay, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

					//write sector header 0x76, sect#, size
					nPos += FMWrite8E1(0x76, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1((unsigned char)(nCount+1), fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1((unsigned char)sect.nBlockLen[nCount], fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					const unsigned char * p = sect.pData[nCount];

					//write sector data
					for (int i = 0; i < sect.nBlockLen[nCount] * 256; i++)
					{
						if (p)
							nPos += FMWrite8E1(*p++, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
						else
							nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					
					}
					//write end of sector marker
					nPos += FMWrite8E1(0x47, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1(0x53, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

					lastlen = sect.nBlockLen[nCount];
					nCount++;


				}
				//write trailing bits or padding
				nPos += FMWriteZeros( (nDiskTargetLen - nPos) * 2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);


			 }
			 else
			 {		//no sectors, unknown or blank formatting

				int nDelay = fpm * lastlen;
				const unsigned char * p = pTrack ? pTrack+4 : NULL; //skip track marker
				nPos += FMWriteZeros( nDelay, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				//write unknown track data, no special formatting
				for (int i = 0; i < (dt == OS65D5 ? SECTORS5 : SECTORS8 ) * 256; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				}

				//write trailing bits or padding
				nPos += FMWriteZeros( (nDiskTargetLen - nPos)*2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);


			}
		}
		else
		{	//OS65U
			//one 3584 byte sector, encoding switches mid-track
			//find start of sector data	within a reasonable about of bytes
			int i = 0;
			bool bFound = false;
			if (pTrack)
			{
				while ( i < 30 && i < nOutLen )	 //
				{
					if (pTrack[i] == 0x80) 
					{
						if (pTrack[i+1] == 0x00)
						{
							if (pTrack[i+2] == nTrack)
							{
								bFound = true;
								i +=3;
								break;
							}
						} 
					}
					i++;
				}
			}
			else
			{
				bFound = true;
			}

			if (bFound)
			{
				if ( pTrack && (pTrack[i] & 0x80) ) //skip possible noise byte, usually $FF, next data usually 0 or 1F
					i++;    
			
				 int chk = 0x80 + nTrack + 0x1F;
				//write leader
				nPos = FMWriteZeros( nDisk8Index, fmd, bitwidth, lpOut, nOutLen);
				//write track header
				nPos += FMWrite8E1(0x80, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				nPos += FMWrite8E1(0x00, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				nPos += FMWrite8E1((unsigned char)nTrack, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				
				//delay after switching encoding -- how much?
				nPos += FMWriteZeros( 44, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
				
				for (int j = 0; j < 0xE03; j++)	 //write sector	+ checksum 
				{
					if (pTrack)
						nPos += FMWrite8N1(pTrack[i+j], fmd, bitwidth, lpOut+nPos, nOutLen-nPos);	
					else
					{
						if (nTrack > 0 && j == 0)
							nPos += FMWrite8N1((unsigned char)0x1F, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
						else if (nTrack > 0 && j == 0xE01)
							nPos += FMWrite8N1((unsigned char) ((chk >> 8) & 0xFF), fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
						else if (nTrack > 0 && j == 0xE02)
							nPos += FMWrite8N1((unsigned char) (chk & 0xFF), fmd, bitwidth, lpOut+nPos, nOutLen-nPos);
						else
							nPos += FMWrite8N1(0, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);	
					}
				}
				//write trailing bits if any
				nPos += FMWriteZeros( (nDiskTargetLen - nPos)*2, fmd, bitwidth, lpOut+nPos, nOutLen-nPos);

			}

		}

	}



	return nPos;
}

#define TRACKBUFSIZE 32768  //size of temp track buffer

//Lets create an HFE file from an OSI disk image.in memory

// pDisk is source OSI disk memory image, or NULL to create blank image
// dt specifies source format and desired target size/type
// sides indicates target # of sides (usually 2)
// ppData is returned HFE disk in memory (caller must free)
// pnLen returns size of HFE image in memory
//
// return true if OK
bool CreateHFEImage( unsigned char * pDisk, DISKTYPE dt, int nSides, int nBitwidth, unsigned char** ppData, int* pnLen, bool bWrite8As5 )
{
	int nTracks = 0; //40 for 5.25, 77 for 8"
	int nTrackLen = 0;
	if (nSides < 1 || nSides > 2)
		nSides = 1;
	if (nBitwidth < 1 || nBitwidth > 2)
		nBitwidth = 2;

	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
		case EMPTY5:
		case EMPTY5Z:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
		case EMPTY8:
		case EMPTY8Z:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}

	unsigned char * pTrackBuf = (unsigned char *) malloc(TRACKBUFSIZE); //buffer for encoding track

	//How big?
	unsigned int dwHeader = 512; //sizeof(picfileformatheader); 
	unsigned int  dwTrackTable =  (sizeof(pictrack) * nTracks);
	 dwTrackTable = ((dwTrackTable + 511) / 512) * 512; //padded table size
	//how big are tracks?
	//5ms index pulse, + data
	//5.25 = 300RPM, or one rotation every 200ms
	//transfer rate is 250Kb
	// max flux per track is 50000 	(25000 bytes)
	// double sided = 100000 /8 = 12500 bytes
	//at 247,000 bps, 1 track can not exceed 49400 bits, 6175 bytes 

	//8" = 360RPM or one rotation every 166.6ms 
	// transfer rate is 500Kb
	// max flux / track is 83300
	// double sided - 166600 /8 = 20825 bytes

	int nTrackTotalSize = 0;
	

	//figure out the size of the encoded tracks
	for (int i = 0; i < nTracks; i++)
	{
		unsigned char * pTrack = NULL;
		if (pDisk)
			pTrack = pDisk + (nTrackLen * i);
		memset(pTrackBuf, 0, TRACKBUFSIZE);
		int nLen = FMEncodeOSITrack( pTrack, dt, i, nBitwidth, pTrackBuf, TRACKBUFSIZE, bWrite8As5);

		//track data is padded
		int nPadded = (nLen + 511) / 512 * 512;
		
		nTrackTotalSize += (nPadded * nSides); //2 = double sided ignored data
	}

	//let make an HFE image;
	unsigned char *pHFE = (unsigned char *) malloc( dwHeader + dwTrackTable + nTrackTotalSize);
	if (pHFE)
	{
		memset(pHFE, 0, dwHeader + dwTrackTable + nTrackTotalSize);
		memset(pHFE, 0xFF, dwHeader+dwTrackTable);


		picfileformatheader* pHead = (picfileformatheader*) pHFE;
		memcpy(pHead->HEADERSIGNATURE, szHeaderSig, sizeof(szHeaderSig));
		pHead->formatrevision = 0;
		pHead->number_of_track = (unsigned char)nTracks;
		pHead->number_of_side = (unsigned char)nSides;
		pHead->track_encoding = 0xFF;
		pHead->bitRate = (unsigned short)(dt == OS65D5 ? 250 : 500);
		pHead->floppyRPM = (unsigned short)(dt == OS65D5 ? 300 : 360);
		pHead->floppyinterfacemode = GENERIC_SHUGGART_DD_FLOPPYMODE;
		pHead->dnu = 1; //(?)
		pHead->track_list_offset = 1; //(=512)
		pHead->write_allowed = 0xFF;
		pHead->single_step = 0xFF;
		pHead->track0s0_altencoding = 0xFF;
		pHead->track0s0_encoding = 0xFF;
		pHead->track0s1_altencoding = 0xFF;
		pHead->track0s1_encoding = 0xFF;

		pictrack* TrkLst = (pictrack*)(pHFE + 0x200);


		unsigned char *pHFETracks = (unsigned char *) (pHFE + dwTrackTable + dwHeader);
		
		//make a list of track lengths
		for (int i = 0; i < nTracks; i++)
		{
			unsigned char * pTrack = NULL;
			if (pDisk)
			{
				pTrack = pDisk + nTrackLen * i; 

			}
			int nLen = FMEncodeOSITrack( pTrack, dt, i,  nBitwidth, pTrackBuf, TRACKBUFSIZE, bWrite8As5);

			memset(pTrackBuf+nLen, 0, TRACKBUFSIZE-nLen); //erase rest of track data 

			//now interleave blocks to create double sided data
			TrkLst[i].offset = (unsigned short) ((pHFETracks - pHFE) / 512);
			TrkLst[i].track_len = (unsigned short) (nLen*nSides);

			int nPadded = (nLen*nSides + 511) / 512 * 512;

			int nBlocks = (nLen + 255) / 256;

			if (nSides == 2) //interleave blocks
			{
				unsigned char *p = pTrackBuf;
				unsigned char *pOut = pHFETracks;
				for (int j = 0; j < nBlocks; j++)
				{
					memcpy(pOut, p, 256);	//copy side 1 block
					pOut += 256;
					p += 256;
					memset(pOut, 0x11, 256); //fill side 2 with FM 0's (blank)
					pOut += 256; 	
					
				}
			}
			else   //single data block
			{
				memcpy(pHFETracks, pTrackBuf, nLen);
			}

			
			pHFETracks += nPadded;
		}

		if (pnLen )
			*pnLen = dwHeader + dwTrackTable + nTrackTotalSize;

		if (ppData)
		{
			*ppData = pHFE;
		}
		else
		{
			free(pHFE);
			pHFE = NULL;
		}
		return true;
	}

	free(pTrackBuf);
	pTrackBuf = NULL;

	return false;
}

//Determine OSI disk type from supplied disk image in memory
// Dos/65 has normal OSI track on track 0, otherwise looking at data is the only way to see
//
DISKTYPE IdentifyDiskType( unsigned char *pMem, int nReadBytes)
{
	DISKTYPE dt = DISK_UNKNOWN;
	if (!pMem || nReadBytes < 2048)
		return dt;
	
	int nTrackLen = TRKSIZ5;

	//is it OSIdisk?  (allow for slop)
	if ( (nReadBytes >  TRKSIZ5 * 41) && (nReadBytes < TRKSIZ8 * 81) )
	{
		nTrackLen = TRKSIZ8;
	}
	if (nTrackLen > nReadBytes)	//too small
		return dt;

	OSISector sect;
	int nFound = 0;
	int i;
	//OS65D 0x43, 0x57, Trk, 0x58, ..., $76, #sec , sectLen,
	//HexDos $53, trk#  

	for (i = 0; i < 20 && nFound < 4; i++) //start at track 1 for a while
	{
		DISKTYPE dSearch = DISK_UNKNOWN;
		int offset = i * nTrackLen;
		unsigned char * pTrack = pMem + i * nTrackLen;

		int nStart =  OSISeekTrackStart( pTrack, nReadBytes-offset, DISK_UNKNOWN, i);

		if (i == 0 && nStart >= 0)
		{
			//look for magic text "RICHARD A. LEARY"
			
			//check track 0	for DOS/65 sector

			unsigned char match[] = {  0x43, 0x57, 0x00, 0x58 };
			if ( !memcmp(pTrack+nStart, &match[0], sizeof(match)/sizeof(match[0])) )
			{ 
				//found OS65D normal track header on trk 0, which is usually reserved as boot trk
				//except on DOS/65
				int nCount = OSICountSectors( pTrack, nTrackLen , i, sect);
				if (nCount == 1)
				{
					if (sect.nBlockLen[0] == 8)
					{
						dt = DOS655;
						nFound = 4;
		
					}
					else if (sect.nBlockLen[0] == 13)
					{
						dt = DOS658;
						nFound = 4;
					}	
				}
			}
		} 
		else if (nStart >= 0)
		{
			if (pTrack[nStart] == 0x43 && pTrack[nStart+1] == 0x57 && 
				pTrack[nStart+2] == (unsigned char) BIN2BCD(i))
			{
				
				int nCount = OSICountSectors( pTrack, nTrackLen , i, sect);

				if (nCount > 0)
				{
					if (nTrackLen == TRKSIZ8)
						dSearch = OS65D8;
					else
						dSearch = OS65D5;
				}
				else if (pTrack[nStart+3] == 0x00)
				{
					//could be CPM 2.2 variant (BCD 0-9 = BINARY 0-9)
					dSearch = OS65D8;
				}
			} 
			else if (pTrack[nStart] == 0x43 && pTrack[nStart+1] == 0x57 && 
					 pTrack[nStart+2] == (unsigned char) i && pTrack[nStart+3] == 0) 
			{
				dSearch = OS65D8;	
			}
			if (pTrack[nStart] == 0x57 && pTrack[nStart+1] == i)
			{
				dSearch = HEXDOS;
			}

			if ( pTrack[nStart] == 0x80 && pTrack[nStart+1] == 0x00 && pTrack[nStart+2] == i )
			{
				dSearch = OS65U8;
			}
			if (dSearch != DISK_UNKNOWN)
			{

				if (dt == dSearch)
				{
					nFound++;
				}
				else if (nFound < 3)
				{
					dt = dSearch;
					nFound = 1;
				}
			}
		}

	}

	return dt;
} 

bool DisplayHFEFileInfo(unsigned char * pHFEDisk, int nHFELen)
{
	if (!pHFEDisk || nHFELen < 1000 )
		return false;
	TCHAR buf[2048];
	const TCHAR *pEncode = _T("");

	picfileformatheader *pHFE = (picfileformatheader*) pHFEDisk;
	switch (pHFE->track_encoding)
	{
		case ISOIBM_MFM_ENCODING:// 0x00
			pEncode = _T("IBM_MFM");
			break;
		case AMIGA_MFM_ENCODING:// 0x01
			pEncode = _T("AMIGA_MFM");
			break;

		case ISOIBM_FM_ENCODING:// 0x02
			pEncode = _T("IBM_FM");
			break;

		case EMU_FM_ENCODING:// 0x03
			pEncode = _T("FM");
			break;

		case UNKNOWN_ENCODING: // 0xFF
			pEncode = _T("RAW");
			break;

		default:
			pEncode = _T("Unexpected value??");
			break;

	}



#if (_MSC_VER >= 1400) && (defined(UNICODE))
			swprintf(buf, _T("HFEFile: %d tracks, %d sides, bitrate:%d, Encoding: %s \r\n"),pHFE->number_of_track, pHFE->number_of_side, pHFE->bitRate, pEncode);
#else							
			_stprintf(buf, _T("HFEFile: %d tracks, %d sides, bitrate:%d, Encoding: %s \r\n"),pHFE->number_of_track, pHFE->number_of_side, pHFE->bitRate, pEncode);
#endif
			if (logfile)
			{
				_fputts( buf, logfile);
			}					
			else
			{
				_fputts(buf, stdout);
			}

	return true;
}

//simply check decoded data for known OSI disk formatting
DISKTYPE DetermineOSIDiskTypeFromHFE(unsigned char * pHFEDisk, int nHFELen, int nSide /*=0*/, int nSkip /*= 1*/)
{
	DISKTYPE dt = DISK_UNKNOWN;
	bool bSingleBit = false;		//flag single bit instead of 2 bit per cell
	int nNumSides = 2;

	if (!pHFEDisk || nHFELen < 1000 || nSide > 2)
		return DISK_ERROR;
	if (nSkip < 1 || nSkip > 4)
		nSkip=1;

	picfileformatheader *pHFE = (picfileformatheader*) pHFEDisk;
	int nTracks = pHFE->number_of_track/nSkip;

	if (nTracks > 80)
	{
		LogText( _T("File has too many tracks (%d)\r\n"), pHFE->number_of_track);

		return DISK_ERROR;

	}
	if (pHFE->track_encoding != UNKNOWN_ENCODING)
	{
		LogText( _T("Warning: HFE file encoding (%02X) is not \"UNKNOWN_ENCODING\"\r\n"), pHFE->track_encoding);

		//return DISK_ERROR;	//no longer an error, just unexpected
		
	}
	if (pHFE->number_of_side < 1 || pHFE->number_of_side > 2)
	{
		LogText( _T("HFE file unexpected number of sides (%d)"),pHFE->number_of_side );

		return DISK_ERROR;

	}
	nNumSides = pHFE->number_of_side == 1 ? 1 : 2;
	pictrack* trkLst = 	(pictrack*) (pHFEDisk + pHFE->track_list_offset * 0x200);

	
	if (pHFE->bitRate <=300 && nTracks > 42)
	{
		bSingleBit = true;
		if (pHFE->number_of_side==1 && trkLst[0].track_len > (DISK8TARGETLEN/2 + DISK8TARGETLEN/10) ) //allow for slower RPM
		{
			//header could be lying about being single sided (greaseweasle 10/2020)
			nNumSides = 2;
		}
			
	} 
	else if (pHFE->bitRate <= 300 && nTracks < 42)
	{
			if (nNumSides == 2 && trkLst[0].track_len < (DISK5TARGETLEN*2 - DISK5TARGETLEN/10) )
				bSingleBit = true;		
	}

	if (nSide > nNumSides )
		nSide = nNumSides;


	FMstate fm;
	void *pEnd = pHFEDisk + nHFELen;
	unsigned char* pTrackBuf = (unsigned char *) malloc (TRKSIZ8);
	if ( !pTrackBuf)
		return DISK_ERROR;
	memset(pTrackBuf, 0, TRKSIZ8); 
	

	if (((void *) &trkLst[0]) < (pEnd))
	{
		int nFound = 0;
		int nCurTrack = 0;

		for (int i = 0; i < pHFE->number_of_track && nFound < 4; i++)	 //lets skip track 0
		{
		    
			
			//does this track have data?
			if (trkLst[i].offset == (unsigned short) -1 || trkLst[i].track_len == (unsigned short) -1)
			{
				LogText( _T("HFE track %d missing...skipping.\r\n"), i);
				continue;
			}
			
			DISKTYPE dSearch = DISK_UNKNOWN;
			unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
			int pos = 0;
			int nDataLen = trkLst[i].track_len;
			int nOutPos = 0;

			int loop = 0;


			//new track, reset decoder
			FMdecodeInitialize(fm);


			while (loop < nDataLen)
			{	
				unsigned char * inBuf = pData + loop;

				//we're reading size 2
				if (nSide == 2)
					loop += 256;

				int nInLen = 256;
				if (nInLen + loop > nDataLen)
				{
					nInLen = nDataLen - loop;
				}

				int nRead = 0;
				int nWrote = 0;

				FMdecodestream(inBuf, nInLen, bSingleBit ? 1 : 2, pTrackBuf+nOutPos, NULL, TRKSIZ8-nOutPos,fm, &nRead, &nWrote, pos+loop, false);		
				
				if (nWrote > 0)
					nOutPos += nWrote;
				if (nWrote >= TRKSIZ8)
				{
					//handle full output buffer
					break;
				}

				if (nSide == 2)
					loop += 256; //add one page to point to side1, updated again above
				else
					//input data is stored in 0x100 chunks, alternating between side 1 & 2
					loop += (nNumSides * 256);

			
			}
			if (nOutPos == 0) //no data encoded on track
			{
				dSearch = nDataLen > (DISK5TARGETLEN*( bSingleBit ? 1 : 2) + 5000) ? EMPTY8 : EMPTY5;
			}

			int offset =  OSISeekTrackStart( pTrackBuf, nOutPos, dt, nCurTrack);
			if (offset >= 0)
			{
				if (pTrackBuf[offset] == 0x43 && pTrackBuf[offset+1] == 0x57)
				{
					if (  pTrackBuf[offset+2] < 0x80 && ( abs(nCurTrack - BCD2BIN( pTrackBuf[offset+2] )) <= 3) )
						nFound++;

					if (nTracks > 42)
						dSearch = OS65D8;
					else
						dSearch = OS65D5;
					if (nCurTrack == 0 && (dt == OS65D8 || dt == OS65D5)) //OSI sector on tk0, probably DOS/65
					{
						OSISector sect;
						memset(&sect, 0, sizeof(OSISector));
						
						if (dSearch == OS65D5 && OSICountSectors( pTrackBuf, nOutPos , nCurTrack, sect) == 1)
						{
							if ( sect.nBlockLen[0] == 8 )  //DOS65 always single sector of 8 or 13 blocks
							{
								dt = DOS655;
								nFound = 4;
							}
							else if (sect.nBlockLen[0] == 13)
							{	
								dt =  DOS658;
								nFound = 4;
							}
						}
					}

					
				}
				if (pTrackBuf[offset] == 0x57 && pTrackBuf[offset+1] == nCurTrack)
				{
					if (pHFE->number_of_track <= 40)
					{
						nFound++;
						dSearch = HEXDOS;
					}

				}
				if ( pTrackBuf[offset] == 0x80 && pTrackBuf[offset+1] == 0x00 && (abs(pTrackBuf[offset+2]-i) <= 3) )
				{
					//we've found OS68U disk type which require different decoding
					
					dSearch = OS65U8;
					nFound++;

					//restart decode of mixed 8E1/8N1
					unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
					int pos = 0;
					int nDataLen = trkLst[i].track_len;
					int nOutPos = 0;

					int loop = 0;


					loop = 0;
					int nRead = 0;
					int nWrote = 0;

					//reset decoder
					FMdecodeInitialize(fm);

					int nLen = 3; //decode 1st 3 bytes
					bool bHeader = true;
					while (loop < nDataLen)
					{	
						unsigned char * inBuf = pData + loop;

						//we're reading size 2
						if (nSide == 2)
							loop += 256;

						int nInLen = 256;
						if (nInLen + loop > nDataLen)
						{
							nInLen = nDataLen - loop;
						}
						if (bHeader)
							nLen = 3 - nOutPos;
						else
							nLen = TRKSIZ8-nOutPos;

						nRead = 0;
						nWrote = 0;

						FMdecodestream(inBuf, nInLen, bSingleBit ? 1 : 2, pTrackBuf+nOutPos, NULL,  nLen,fm, &nRead, &nWrote, pos+loop, false);		

						if (nWrote > 0)
							nOutPos += nWrote;
						if (nWrote >= nLen)
						{
							if (bHeader)
							{
								FMdecodeInitialize(fm);
								fm.eSformat = s8N1;
								bHeader = false;
								nLen = TRKSIZ8-nOutPos;
								//decode rest of input buffer
								FMdecodestream(inBuf+nRead, nInLen-nRead, bSingleBit ? 1 : 2, pTrackBuf+nOutPos, NULL,  nLen,fm, &nRead, &nWrote, pos+loop, false);		

							}
							else
							{
								//handle full output buffer
								break;

							}
						}

						if (nSide == 2)
							loop += 256; //add one page to point to side1, updated again above
						else
							//input data is stored in 0x100 chunks, alternating between side 1 & 2
							loop += (nNumSides * 256 );

					
					}

				}
			}
			if (dSearch != DISK_UNKNOWN)
			{

				if (dt == dSearch)
				{
					nFound++;
				}
				else if (nFound < 3)
				{
					dt = dSearch;
					nFound = 1;
				}
			}

			i += nSkip-1;
			nCurTrack++;
		}


	}	
	
	
	free(pTrackBuf);

	
	return dt;	
}

//convert HFE file to OSI Disk
#define FLOPPY_DATA_SIZE 0x60000

//Lets create an OSI dump image from an OSI HFE disk image.in memory
//
// Decode an HFE image, emulating OSI decoding hardware, determines target type from source HFE image
// pHFEDisk - memory image of HFE file
// nHFELen - and its size
// nSide  - source side (0 or 1/ or 2)
// ppOSI - returned OSI memory image - caller must free
// pnOSILen - and its size
bool CreateOSIImage( unsigned char * pHFEDisk, int nHFELen, int nSide, int nSkip, unsigned char** ppOSI, int* pnOSILen, DISKTYPE* pdt, bool bVerbose )
{

	unsigned char * OSIDisk = NULL;
	unsigned char * ErrDisk = NULL;
	int ErrorOffset = 0;
	pictrack* trkLst = NULL;
	picfileformatheader *pHFE = (picfileformatheader*) pHFEDisk;;
	unsigned char * pEnd = pHFEDisk + nHFELen;
	FMstate fm;
	int nTrackSize = TRKSIZ5;
	DISKTYPE dt = DISK_UNKNOWN;
	bool bSinglebit = false;
	int nNumTracks = 0;

	if (pdt)
		*pdt = dt;
	int nTrkOffset = 0;

	if (nSide < 0 || nSide > 2)
		return false;

	if (nSide == 0)
		nSide = 1;

	int nNumSides = 2;
	if (!pHFEDisk || !ppOSI)
		return false;

	if (nSkip < 1 || nSkip > 4)
		nSkip = 1;


	if (memcmp(pHFEDisk, szHeaderSig, 8))
	{
		//invalid data
		LogText( _T("HFE file header not found in input file"));
		goto abort;
	}

	nNumTracks = pHFE->number_of_track / nSkip;

	//greaseweazle is setting encoding based on scan parameters (incorrectly identifying as say C64 dd) so ignore this condition
	if (nNumTracks > 80 || /*pHFE->track_encoding != UNKNOWN_ENCODING ||*/ pHFE->number_of_side < 1 || pHFE->number_of_side > 2)
	{
		if (nNumTracks > 80)
		{
			LogText( _T("Unsupported HFE file type, too many tracks"));
			goto abort;
		}
		else
		{
			pHFE->number_of_track = 80;  //ignore long tracks
			nNumTracks = 80/nSkip;
		}

	}
	if (nSide > nNumSides )
		nSide = nNumSides;



	if (nNumTracks > 40)
		nTrackSize = TRKSIZ8;

	//allocate memory for OSIDisk image and decode error data
	OSIDisk = (unsigned char *) malloc (nNumTracks * nTrackSize *2);
	if (!OSIDisk)
	{
		LogText( _T("Error allocating target disk image memory"));
		goto abort;
	}

	memset(OSIDisk, 0, nNumTracks * nTrackSize *2);
	ErrorOffset = nNumTracks * nTrackSize;
	ErrDisk = OSIDisk + ErrorOffset;
	trkLst = (pictrack*) (pHFEDisk + pHFE->track_list_offset * 0x200);
	nNumSides = pHFE->number_of_side == 1 ? 1 : 2;
	if (pHFE->bitRate <=300 && nNumTracks > 42)
	{
		bSinglebit = true;
		if (pHFE->number_of_side==1 && trkLst[0].track_len > (DISK8TARGETLEN/2 + DISK8TARGETLEN/10) )
		{
			//header could be lying about being single sided (greaseweasle 10/2020)
			nNumSides = 2;
		}
	}

	if ((unsigned char *) &trkLst[0] < pEnd)
	{
		int nCurTrack = 0;
		
		for (int i = 0; i < pHFE->number_of_track; i++)
		{

			//does this track have data?
			if (trkLst[i].offset == (unsigned short) -1 || trkLst[i].track_len == (unsigned short) -1)
			{
				if (bVerbose)
				LogText( _T("Error -Track %d is missing from data file...skipping\r\n"), i);
				
				continue;
			}
			
			
			unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
			int pos = (pData - pHFEDisk);
			int nDataLen = trkLst[i].track_len;
			int nOutPos = 0;

			int loop = 0;
			if (bVerbose)
				LogText( _T("\r\nTrack %d @ $%04lX-$%04lX ($%04lX long)\r\n--------\r\n"),i,trkLst[i].offset * 0x0200, trkLst[i].offset * 0x0200 + trkLst[i].track_len, trkLst[i].track_len);

			//new track, reset decoder
			FMdecodeInitialize(fm);
			unsigned char * pOut = OSIDisk + i * nTrackSize;
			unsigned char * pErr = pOut + ErrorOffset;

			while (loop < nDataLen)
			{
				unsigned char * inBuf = pData + loop;
				int nInLen = 256;
				if (nInLen + loop > nDataLen)
				{
					nInLen = nDataLen - loop;
				}
				pOut = OSIDisk + nCurTrack * nTrackSize + nOutPos;
				pErr = pOut + ErrorOffset;
				int nOutSize = nTrackSize;
				int nRead = 0;
				int nWrote = 0;

				FMdecodestream(inBuf, nInLen, bSinglebit ? 1 : 2, pOut, pErr, nOutSize,fm, &nRead, &nWrote, pos+loop, bVerbose);

				if (nWrote > 0)
					nOutPos += nWrote;

				if (nWrote >= nOutSize)
				{
					//handle full output buffer
				}

				//input data is stored in 0x100 chunks, alternating between side 1 & 2
				//we only want 1 side's data
				loop += (nNumSides * 256   );
			
			}
			if (fm.nZeroRun > 20)
			{
				if (bVerbose)
					LogText( _T("\r\n Zeros %02d \r\n"),fm.nZeroRun);
			}
			if (dt == DISK_UNKNOWN )
			{
				pOut = OSIDisk + nCurTrack * nTrackSize;
				pErr = pOut + ErrorOffset;
				int offset =  OSISeekTrackStart( pOut, nOutPos);
				if (offset >= 0)
				{
					if (pOut[offset] == 0x43 && pOut[offset+1] == 0x57)
					{
						if (pOut[offset+2] < 0x80)
						{
							int nTrk = BCD2BIN(	pOut[offset+2] );
							if (abs(nTrk - nCurTrack) <= 3)
							{
								nTrkOffset = nCurTrack-nTrk;
								if (nNumTracks > 40)
									dt = OS65D8;
								else
									dt = OS65D5;
							}
							if (nCurTrack == 0 && (dt == OS65D8 || dt == OS65D5)) //OSI sector on tk0, probably DOS/65
							{
								OSISector sect;
								memset(&sect, 0, sizeof(OSISector));
								
								if (dt == OS65D5 && OSICountSectors( pOut, nOutPos , nCurTrack, sect) == 1)
								{
									if ( sect.nBlockLen[0] == 8 )  //DOS65 always single sector of 8 or 13 blocks
										dt = DOS655;
									else if (sect.nBlockLen[0] == 13)
										dt =  DOS658;
								}
							}
						}
					}
					else if ( pOut[offset] == 0x57 && pOut[offset+1] == nCurTrack )
					{
						if (nNumTracks <= 40)
							dt = HEXDOS;
					}

					else if ( pOut[offset] == 0x80 && pOut[offset+1] == 0x00  )
					{
						//we've found OS68U disk type which require different decoding
						int nTrk = pOut[offset+2];
						if (abs(nTrk - nCurTrack) <= 3)
						{
							nTrkOffset = nCurTrack-nTrk;

							dt = OS65U8;
						}
					}
				}



			}
			if (nCurTrack > 0 && dt == OS65U8)
			{
				//we have OS68U disk type which require different decoding, start over
				//restart decode of mixed 8E1/8N1
				FMdecodeInitialize(fm);

				bool bHeader = true;
				loop = 0;
				nOutPos = 0;
				pData = pHFEDisk + (trkLst[i].offset * 0x0200);
				pos = (pData - pHFEDisk);
				nDataLen = trkLst[i].track_len;
				memset(OSIDisk + nCurTrack * nTrackSize, 0, nTrackSize);

				while (loop < nDataLen)
				{	

					unsigned char * inBuf = pData + loop;
					int nInLen = 256;	 

					if (nInLen + loop > nDataLen)
					{
						nInLen = nDataLen - loop;
					}

					pOut = OSIDisk + nCurTrack * nTrackSize + nOutPos;
					pErr = pOut + ErrorOffset;
					int nOutSize = bHeader ? 3 : nTrackSize; //decode 1st 3 bytes
					int nRead = 0;
					int nWrote = 0;

					FMdecodestream(inBuf, nInLen, bSinglebit ? 1 : 2, pOut, pErr, nOutSize-nOutPos,fm, &nRead, &nWrote, pos+loop, bVerbose);

					if (nWrote > 0)
						nOutPos += nWrote;
					
					if ( nOutPos >= nOutSize)
					{
						if (bHeader)
						{
							FMdecodeInitialize(fm);
							fm.eSformat = s8N1;

							bHeader = false;
							nOutSize = nTrackSize-nOutPos;
							pOut += nWrote;	
							pErr = pOut + ErrorOffset;
							nWrote = 0;
							inBuf += nRead;
							
							//decode rest of input buffer
							FMdecodestream(inBuf, nInLen-nRead, bSinglebit ? 1 : 2,  pOut, pErr,  nOutSize,fm, &nRead, &nWrote, pos+loop+nRead, false);		
							if (nWrote > 0)
								nOutPos += nWrote;
						}
						else
						{
							//handle full output buffer
							break;

						}
					}

					//input data is stored in 0x100 chunks, alternating between side 1 & 2
					//we only want 1 side's data
					loop += (nNumSides * 256   );
				
				} //while decoding this track

			} //if OS65U

			i+=nSkip-1;
			nCurTrack++;
		
		} //for there are tracks

	}
	if (nTrkOffset > 0)
	{
			if (bVerbose)
				LogText( _T("Track Data skewed by %d tracks\r\n"),nTrkOffset);
		
		memmove(OSIDisk, OSIDisk+(nTrkOffset*nTrackSize), (pHFE->number_of_track-nTrkOffset)*nTrackSize);
	}

	if (ppOSI)																		   
		*ppOSI = OSIDisk;
	else
		free (OSIDisk);
	
	if (pdt)
		*pdt = dt;

	if (pnOSILen)
	{
		int nSize = nNumTracks;
		if (nSize > 77)
			nSize = 77;
		*pnOSILen = nSize * nTrackSize;
	}

	return true;
abort:

	if (OSIDisk)
	{
		free(OSIDisk);
		OSIDisk = NULL;
	}
	if (ppOSI)
		*ppOSI = NULL;
	if (pnOSILen)
		*pnOSILen = -1;

	return false;
}

//Lets clean an OSI disk Image in memory

// pDisk is source OSI disk memory image
// dt specifies source format and desired target size/type
// 
// ppData is returned OSI disk in memory (caller must free)
// pnLen returns size of OSI image in memory
//
// return true if OK
bool CleanOSIImage( unsigned char * pDisk, DISKTYPE dt, unsigned char** ppData, int* pnLen )
{
	int nTracks = 0; //40 for 5.25, 77 for 8"
	int nTrackLen = 0;


	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}

	unsigned char * pTrackBuf = (unsigned char *) malloc(nTrackLen * nTracks); //buffer for encoding track

	if (!pTrackBuf)
		return false;
	memset(pTrackBuf, 0, nTrackLen * nTracks); //zero track data
	  
	if (pnLen)
		*pnLen = nTrackLen * nTracks;
	if (ppData)
		*ppData = pDisk;
	
	if (pDisk)
	{
		int nTrack;
		OSISector sect;
		memset(&sect, 0, sizeof(OSISector));
		bool bDos65Tk0 = false;
		
		if (dt == OS65D5 && OSICountSectors( pDisk, nTrackLen , 0, sect) > 0)
		{
			//normal OS65D disks do not have a regular track/sector on tk0
			bDos65Tk0 = true;
			if (dt == OS65D5) dt = DOS655;
			if (dt == OS65D8) dt = DOS658;
		}
		LogText( _T("Clean: disk type %s\r\n"),  szIdentity[dt]);


		for ( nTrack = 0; nTrack < nTracks; nTrack++)
		{
			unsigned char * pDestTrack = pTrackBuf + nTrack * nTrackLen;
			unsigned char * pDestTrackStart = pDestTrack;
			unsigned char * pSrcTrack = pDisk + nTrack * nTrackLen;

			int nSectors = OSICountSectors( pSrcTrack, nTrackLen , nTrack, sect);
			if (nTrack == 0 && nSectors == 0)
			{
				//just copy whole track, some OS use 8,9 or 13 blocks
				//note hexdos allows for ~170 byte autostart program at the end of (only) trk 0 boot ($800 bytes)

				memcpy(pDestTrack, pSrcTrack, nTrackLen);				
			}
			else
			{
				int nChk = 	(*(pSrcTrack+nTrackLen-3) << 8) | *(pSrcTrack+nTrackLen-4);					
				int nMax =  (*(pSrcTrack+nTrackLen-1) << 8) | *(pSrcTrack+nTrackLen-2);
				if ((nChk ^ 0xFFFF) != nMax)
					nMax = -1;
				else
				{
					int i;
					for (i = nTrackLen-4; i < nTrackLen; i++)
						*(pSrcTrack+i) = 0; //zero out EOT marker
				}				
				
				
				if (dt == HEXDOS)
				{
					//write track id
					*pDestTrack++ = 0x57;
					*pDestTrack++ = (unsigned char) nTrack;
					 
					if (nSectors > 0) 
					{
						const unsigned char * p = sect.pData[0];

						//write sector data
						for (int j = 0; j < sect.nBlockLen[0]; j++)
						{
							*pDestTrack++ =*p++;				
						}

					 }
					 else
					 {		//no sectors, unknown or blank formatting

						const unsigned char * p = pSrcTrack ? pSrcTrack+2 : NULL; //skip track marker

						//write unknown track data, no special formatting
						for (int j = 0; j <  2048; j++)
						{
							*pDestTrack++ =*p++;

						}

					 }

				}
				else if (dt == OS65D5 || dt == OS65D8 || dt == DOS655 || dt == DOS658)
				{
					int nCount = 0;
					
					//write track leader
					*pDestTrack++ = 0x43;
					*pDestTrack++ = 0x57;
					*pDestTrack++ = (unsigned char)BIN2BCD(nTrack);
					*pDestTrack++ = 0x58;

	
					if (nSectors > 0) 
					{
						for (nCount = 0; nCount < nSectors; nCount++)
						{
						
							//write sector header
							*pDestTrack++ = 0x76;
							
							if (sect.nBlockLen[nCount] < 0)
							{
								//write broken sector data from DOS65
								const unsigned char * p = sect.pData[nCount];

								//write sector data
								for (int j = 0; j < -sect.nBlockLen[nCount]; j++)
								{
									*pDestTrack++ = *p++;
								
								}

							}
							else
							{
							
								*pDestTrack++ = (unsigned char)(nCount+1);
								*pDestTrack++ = (unsigned char)sect.nBlockLen[nCount];
								const unsigned char * p = sect.pData[nCount];

								//write sector data
								for (int j = 0; j < sect.nBlockLen[nCount] * 256; j++)
								{
									*pDestTrack++ = *p++;
								
								}
								//write end of sector marker
								*pDestTrack++ = 0x47;
								*pDestTrack++ = 0x53;
							}


						}

					}
					else 
					{	
						//no sectors? just copy entire remaining track data
						int i = OSISeekTrackStart(pSrcTrack, nTrackLen);
						if (i >= 0)	//index of $43
						{
							i+=4; //skip track header
							memcpy(pDestTrack, pSrcTrack+i, nTrackLen-i);
							LogText( _T("Trk %02d Empty Track (0 sectors)\r\n"),  nTrack);

						}
						else
						{
							LogText( _T("Trk %02d : No Track Header\r\n"),  nTrack);

							memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
						}
						
					}
				}
				else if (dt == OS65U8)
				{	//OS65U
					//one 3584 byte sector, encoding switches mid-track
					//find start of sector data	within a reasonable about of bytes
					int j = OSISeekTrackStart(pSrcTrack, nTrackLen, dt);

					if (j >= 0 && pSrcTrack[j] == 0x80)
					{
						j +=3; //skip track start

						if ( pSrcTrack && (pSrcTrack[j] & 0x80) ) //skip possible noise byte, usually $FF, next data usually 0 or 1F
							j++;    
					
						//write track header
						*pDestTrack++ = 0x80;
						*pDestTrack++ = 0x00;
						*pDestTrack++ = (unsigned char)nTrack;
						memcpy(pDestTrack, pSrcTrack+j, 0xE03);	//sector data + chksum
						int chksum = 0;
						int i;
						for (i = 0; i <= 0xE03; i++)
						{
							//sum including track header
							unsigned char c = pDestTrack[i-3];
							chksum += c;						

						}
						if (pDestTrack[0xE02] != (chksum & 0xFF)  || pDestTrack[0xE01] != ((chksum >> 8) & 0xFF))
						{
							LogText( _T("Trk %02d Chksum invalid (patched)\r\n"),  nTrack);
							pDestTrack[0xE02] = (unsigned char) (chksum & 0xFF);
							pDestTrack[0xE01] = (unsigned char)((chksum >> 8) & 0xFF);
						}
						pDestTrack+= 0xE03;
						//


					}
					else
					{
						LogText( _T("No Track Header on track: %02d\r\n"),  nTrack);
						memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
						nMax = -1;
					}
				}
				else
				{	//unknown disk format?
					memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
				}

				if (nMax > 0)
				{
					int nLen = pDestTrack-pDestTrackStart;
					if (nLen > nMax)
					{
						nLen = nMax;	
					}
					int nChk = (nLen ^ 0xFFFF) & 0xFFFF;
					 *(pDestTrackStart+nTrackLen-3) = (unsigned char)((nChk >> 8) & 0xFF);
					 *(pDestTrackStart+nTrackLen-4) = (unsigned char)(nChk & 0xFF);
					 *(pDestTrackStart+nTrackLen-1) = (unsigned char)((nLen >>8) & 0xFF);
					 *(pDestTrackStart+nTrackLen-2) = (unsigned char)(nLen & 0xFF);


					}

			}

		} //for tracks
		if (ppData)
		{
			*ppData = pTrackBuf;
		}
		else
		{
			if (ppData)
				*ppData = NULL;
			if (pnLen)
				*pnLen = 0;

			free(pTrackBuf);
		}

		return true;

	}
	else
	{
		//create empty disk
		int nTrack;
		for ( nTrack = 0; nTrack < nTracks; nTrack++)
		{
			unsigned char * pDestTrack = pTrackBuf + nTrack * nTrackLen;			
			if (dt == HEXDOS)
			{
				//write track id
				*pDestTrack++ = 0x57;
				*pDestTrack++ = (unsigned char) nTrack;
				 
			}
			else if (dt == OS65D5 || dt == OS65D8 || dt == DOS655 || dt == DOS658)
			{
				
				if (nTrack > 0 || (dt == DOS655 || dt == DOS658))
				{
					//write track leader
					*pDestTrack++ = 0x43;
					*pDestTrack++ = 0x57;
					*pDestTrack++ = (unsigned char)BIN2BCD(nTrack);
					*pDestTrack++ = 0x58;
				}
			}
			else if (dt == OS65U8)
			{	//OS65U

				//write track header
				*pDestTrack = 0x80;
				*(pDestTrack+1) = 0x00;
				*(pDestTrack+2) = (unsigned char)nTrack;
				*(pDestTrack+3)	= 0x1F;

				int chksum =0, i;
				for (i = 0; i <= 0xE03; i++)
				{
					//sum including track header
					unsigned char c = pDestTrack[i];
					chksum += c;						

				}
				pDestTrack[0xE05] = (unsigned char)(chksum & 0xFF);
				pDestTrack[0xE04] = (unsigned char)((chksum >> 8) & 0xFF);


			}
		}
	}//empty disk

	return true;


}

//given a sector error, flag the corresponding file in DirMap
//
int AddSectorError(int nTrack, int nOffset, DISKTYPE dt, OSIDirMap* pDirMap)
{
	int nTracks, nTrackLen;
	int nReturn = -1;
	if (!pDirMap)
		return false;
	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}



	if (dt == OS65U8)
	{

	}
	else
	{	//HexDos or OS65D
		
		if (pDirMap)
		{

			int nBase = nTrack * nTrackLen / 128;
			nBase += (nOffset/128);
			
			int idx = pDirMap->BAM128[nBase];
			if ( idx < 64)
			{
				pDirMap->Errors[idx]++;
				nReturn = 1;
			}

		}


	}

	return nReturn;

}

//mark whole track in error, affecting all files stored on this track
bool AddTrackError(int nTrack, DISKTYPE dt, OSIDirMap* pDirMap)
{
	int nTracks, nTrackLen;
	bool bAffected = false;
	if (!pDirMap)
		return false;
	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}
	if (pDirMap)
	{
		
		int nBase = nTrack * nTrackLen / 128;
		int nEnd = ((nTrack+1) * nTrackLen -1) / 128;
		int i;
		for (i = nBase; i < nEnd; i++)
		{
			int idx = pDirMap->BAM128[i];
			if ( idx < 64)
			{
				pDirMap->Errors[idx]++;
				bAffected = true;
			}
		}

	}
	return bAffected;
}

// Lets test an OSI image in memory
//
// Test track formatting, 
//	+OS65U - test track CRC
//

//

// pDisk is source OSI disk memory image
// dt specifies source format and desired target size/type
// 
// ppData is returned OSI disk in memory (caller must free)
// pnLen returns size of OSI image in memory
//
// return 1 if OK
//        0 fail
//        <0 # errors
int TestOSIImage( unsigned char * pMem, DISKTYPE dt, bool bErrorTrackPresent )
{
	int nTracks = 0; //40 for 5.25, 77 for 8"
	int nTrackLen = 0;
	int nTrackError = 0;
	int nReturn = 0;

	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return 0;
	}
 


	int nTrack;
	OSISector sect;
	memset(&sect, 0, sizeof(OSISector));

	//OSIDirMap is used to map disk errors to files
	OSIDirMap dirMap;
	memset(&dirMap, 0, sizeof(OSIDirMap));
	bool bDos65Tk0 = false;


	TCHAR * pText = GetOSIDirectory(pMem, dt, &dirMap);
	if (pText)
	{
		free(pText);
		pText = NULL;
	}

	
	if (dt == OS65D5 && OSICountSectors( pMem, nTrackLen , 0, sect) > 0)
	{
		//normal OS65D disks do not have a regular track/sector on tk0
		bDos65Tk0 = true;
		if (dt == OS65D5) dt = DOS655;
		if (dt == OS65D8) dt = DOS658;
	}
	LogText( _T("Test: disk type %s\r\n"),  szIdentity[dt]);
	int nErrorOffset = nTrackLen * nTracks;

	for ( nTrack = 0; nTrack < nTracks; nTrack++)
	{

		const unsigned char * pSrcTrack = pMem + nTrack * nTrackLen;
		const unsigned char * pErrTrack = bErrorTrackPresent ? pSrcTrack + nErrorOffset : NULL;
		
		int nSectors = OSICountSectors( pSrcTrack, nTrackLen , nTrack, sect);
		nTrackError = 0;

		if (nTrack == 0 && nSectors == 0)
		{
			// OS65D: loadHi,loadLo,#pages
			int load = (*pSrcTrack << 8) | *(pSrcTrack+1);
			int len = *(pSrcTrack+2) * 256;

			if (len > nTrackLen)
			{
				LogText( _T("Trk 00 invalid bootable floppy page length:%02d (Possible HD image?)\r\n"),  *(pSrcTrack+2));
				len = nTrackLen;
				nTrackError++;
	
			}
			else if (load == 0 || len == 0)
			{
				LogText( _T("Trk 00 Not valid boot format\r\n"));	
			}
			else
			{
				LogText( _T("Trk 00 Boot @ $%04X/%d pages\r\n"), load, *(pSrcTrack+2) );
			}

			if (pErrTrack)
			{
				int i;
				for (i = 0; i < len+3; i++)
				{
					if ( *(pErrTrack+i) != 0 )
					{
						LogText( _T("    bit error offset:%02d\r\n"),  i);
						nTrackError++;
					}
				}

			}
			
					
		}
		else
		{
		 
			if (dt == HEXDOS)
			{
				int i = OSISeekTrackStart( pSrcTrack, 32);
				if (i >= 0)
				{
					if ( pSrcTrack[i+1] != (unsigned char) nTrack)
					{
						LogText( _T("Trk %02d invalid track#(%d) in track header\r\n"),  nTrack, (const int) pSrcTrack[i+1]);
						nTrackError++;
					}
				}

				LogText( _T("Trk %02d "),nTrack );

				if (nSectors > 0) 
				{
					const unsigned char * p = sect.pData[0];
					pErrTrack = bErrorTrackPresent ? p + nErrorOffset : NULL;

					LogText( _T("%d/%d \r\n"),1, sect.nBlockLen[0]/256 );
					
					//write sector data
					for (int j = 0; j < sect.nBlockLen[0]; j++)
					{
					
						if (pErrTrack)
						{
							if (pErrTrack+j != 0)
							{
								LogText( _T("      Sector 1 bit error offset:%02d\r\n"), j);
								AddSectorError(nTrack, j, dt, &dirMap);
								nTrackError++;
							}
						}
					}

				 }
				 else
				 {		//no sectors, unknown or blank formatting

					LogText( _T("no header\r\n"));
					int nSecErr = 0;
					const unsigned char * p = pSrcTrack ? pSrcTrack+2 : NULL; //skip track marker
					pErrTrack = bErrorTrackPresent ? p + nErrorOffset : NULL;

					//write unknown track data, no special formatting
					for (int j = 0; p != NULL && j <  2048; j++)
					{
						if (pErrTrack && pErrTrack[j] != 0)
						{
							if (nSecErr == 0)
								LogText( _T("     bit error offset:%02d\r\n"), j);
							nSecErr++;
						}							

					}
					if (nSecErr)
						nTrackError++;	
					if (nSecErr > 1)
						LogText( _T("(%d more errors)"), nSecErr-1);
					if (nSecErr)
						LogText( _T("\r\n"));


				 }
			}
			else if (dt == OS65D5 || dt == OS65D8 || dt == DOS655 || dt == DOS658)
			{
				int nCount = 0;
				const unsigned char *p = pSrcTrack;
				bool bLog = false;

				int i = OSISeekTrackStart( pSrcTrack, 32, dt,nTrack );

				if (i >= 0 && *(p+i+2) != (unsigned char)BIN2BCD(nTrack))
				{
					LogText( _T("Trk %02d invalid track#(%d) in track header. "),  nTrack, *(p+i+2));
					nTrackError++;
					bLog = true;	
				}


				if (nSectors > 0) 
				{
					int nSectLen = 0;
					int nCurSect = 1;
					int nSectPos = 0;
					bool blogged = false;

					if (!bLog)
						LogText( _T("Trk %02d "),nTrack );
					for (nCount = 0; nCount < nSectors; nCount++)
					{
						if (sect.nBlockLen[nCount] < 0)
						{
						}
						else
						{
							nSectLen += sect.nBlockLen[nCount] *256;
							const unsigned char *p = sect.pData[nCount];
							unsigned char c = *(p-2);
							if (  (c >> 4) > 9 || (c & 0x0F) > 9)
							{
								//invalid sector #
								LogText( _T("(invalid sect #%d) "), (int) c);
							}
							else if (BCD2BIN(c) != nCurSect)
							{
								int i = BCD2BIN(c);
								if (i != nCurSect && !blogged)
								{
									blogged = true;
									if (i+1 > nCurSect)
										LogText( _T("(missing sector #%d) "), nCurSect);
									else
									   LogText( _T("(unexpected sector #%d) "), i);

								}
							}


							LogText( _T("%d/%d "),nCount+1, sect.nBlockLen[nCount] );
						}
						nCurSect++;
					}
					LogText( _T("\r\n") );

					
					
					for (nCount = 0; nCount < nSectors; nCount++)
					{
					
						if (sect.nBlockLen[nCount] < 0)
						{

							//check sector data
							for (int j = 0; j < -sect.nBlockLen[nCount]; j++)
							{
								nSectPos++;

								if (pErrTrack)
								{
									if (*(pErrTrack+j) != 0)
									{
										LogText( _T("Trk %02d Sector %d bit error offset:%02d\r\n"), nTrack, nCount+1, j);
										AddSectorError(nTrack, nSectPos, dt, &dirMap);
										nTrackError++;	
									}
								}
							
							}

						}
						else
						{
						
							p = sect.pData[nCount];
							pErrTrack = bErrorTrackPresent ? p + nErrorOffset : NULL;

							int nSecErr = 0;

							//process sector data
							for (int j = 0; j < sect.nBlockLen[nCount] * 256; j++)
							{
								nSectPos++;
								if (pErrTrack)
								{
									if (*(pErrTrack+j) != 0)
									{
										if (nSecErr == 0)
											LogText( _T("      Sector %d bit error offset:%02d "), nCount+1, j);
										AddSectorError(nTrack, nSectPos, dt, &dirMap);
										nSecErr++;

									}
								}
							
							}
							if (nSecErr)
								nTrackError++;	
							if (nSecErr > 1)
								LogText( _T("(%d more errors)"), nSecErr-1);
							if (nSecErr)
								LogText( _T("\r\n"));

							//check end of sector marker? 0x47, 0x53

						}


					}

				}
				else 
				{	
					//no sectors? 
					int i = OSISeekTrackStart(pSrcTrack, nTrackLen);
					if (i >= 0)	//index of $43
					{
						i+=4; //skip track header
						LogText( _T("Trk %02d No Sector 1\r\n"),  nTrack);
						AddTrackError(nTrack, dt, &dirMap);
						nTrackError++;

					}
					else
					{
						LogText( _T("Trk %02d : No Track Header\r\n"),  nTrack);
						AddTrackError(nTrack, dt, &dirMap);
						nTrackError++;
					}
					
				}
			}
			else if (dt == OS65U8)
			{	//OS65U
				//one 3584 byte sector, encoding switches mid-track
				//find start of sector data	within a reasonable about of bytes
				int j = OSISeekTrackStart(pSrcTrack, nTrackLen, dt);

				if (j >= 0 && pSrcTrack[j] == 0x80)
				{
					
					if (pSrcTrack[j+2] != (unsigned char)nTrack)
					{
						LogText( _T("Trk %02d invalid track#(%d) in track header\r\n"),  nTrack, pSrcTrack[j+2]);
						nTrackError++;
					}
					j +=3; //skip track start

					if ( pSrcTrack && (pSrcTrack[j] & 0x80) ) //skip possible noise byte, usually $FF, next data usually 0 or 1F
						j++;    
				
					unsigned char tbuff[0xe08];
					unsigned char *pDestTrack = tbuff;
					//write track header
					*pDestTrack++ = 0x80;
					*pDestTrack++ = 0x00;
					*pDestTrack++ = (unsigned char)nTrack;
					memcpy(pDestTrack, pSrcTrack+j, 0xE03);	//sector data + chksum
					int chksum = 0;
					int i;
					for (i = 0; i <= 0xE03; i++)
					{
						//sum including track header
						unsigned char c = pDestTrack[i-3];
						chksum += c;						

					}
					if (pDestTrack[0xE02] != (chksum & 0xFF)  || pDestTrack[0xE01] != ((chksum >> 8) & 0xFF))
					{
						LogText( _T("Trk %02d Checksum invalid \r\n"),  nTrack);
					    AddTrackError(nTrack, dt, &dirMap);
						nTrackError++;

					}
					//


				}
				else
				{
					LogText( _T("Trk %02d No Track Header\r\n"),  nTrack);
					
				}

			}
			else
			{	//unknown disk format?
				LogText( _T("Trk %02d Unknown Track format\r\n"),  nTrack);
			}
		}

	}
	int i;
	for (i=0; i < 63; i++)
	{
		TCHAR name[63];

		if (dirMap.Errors[i] > 0)
		{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
							swprintf(name, _T("%S"), dirMap.Name[i] );
#elif (defined(UNICODE))							
							_stprintf(name,  _T("%S"), dirMap.Name[i] );
#else
							sprintf(name,  _T("%s"), dirMap.Name[i] );
#endif

			
			LogText( _T("Possibly damaged file %s\r\n"), name);
		}
	}
	if (nTrackError)
		nReturn =-nTrackError;
	else
		nReturn = 1;

	return nReturn;



}

//Read sequential CPM records of 128 bytes each	return #bytes read or -trk# if error. pData should be allocated to hold
// (nRecEnd - nRecStart + 1) *128 bytes, if NULL returns # bytes needed
//performs deinterleave of records
int ReadCPMRecords(unsigned char * pMem, DISKTYPE dt, int nRecStart, int nRecEnd, int nInterleave /*=-1*/,unsigned char *pData)
{
	if (pMem)
	{

		//set disk parameters 5.25"
		int nTrackLen = TRKSIZ5;
		int nBAMlo =  DOS65DIRTRACK5;
		int nTrackRec = 16;  //5.25" DOS65
		int nMaxBlk = 71;
		int nTracks = 40;


		if (dt == DOS658 || dt == DOS655)
		{
			nInterleave = 1;
		}
		else
		{
			if (nInterleave < 0)
				nInterleave = 6; //blk interleave default OSI CPM 2.2, 1.4
		}

		
		if (dt == DOS658 || dt == OS65D8 )
		{
			//set disk parameters 8"
			nTrackLen = TRKSIZ8;
			nBAMlo =  DOS65DIRTRACK8;
			nTrackRec = 26;
			nMaxBlk = 239;
			nTracks = 77;
		}

	
		int nLastTrk = -1;
		int n = 0;
		int nDataLen = 0;
		OSISector S;
		unsigned char *pCur = NULL;
		memset(&S, 0, sizeof(OSISector));


		if (pData)
			pCur = pData;

		for (int rec = nRecStart; rec <= nRecEnd; rec++)
		{
		
			//convert blk# to record #
			int trk = (rec/nTrackRec);  //rec# / recs per track = trk
			int remainder = rec - (trk * nTrackRec); //remainder = offset to record
			trk += nBAMlo; //set base offset

			unsigned char* pTrack = pMem + trk*nTrackLen;
			
			if (trk != nLastTrk)
			{
				memset(&S, 0, sizeof(OSISector));
				n = OSICountSectors(pTrack, nTrackLen, trk, S);
				if (n < 1)
				{
					//error reading track!
					return -trk;
				}
				nLastTrk = trk;
			}

			//CPM blocks can be (are) interleaved by 6
			int off = (remainder * nInterleave) % nTrackRec; 
			if (nInterleave > 1 && remainder >12) 
				off++;

			const unsigned char *pDat = S.pData[0] + off * 0x80; //offset block * sizeof record
			//128bytes of data
			if (pCur)
			{
				memcpy(pCur, pDat, 128);
				pCur = pCur + 128;
			}
			nDataLen+= 128;
			
		}
	
		return nDataLen;

	}

	return -1;


}

//Retrieves CPM or DOS65 directory, embeddded in OS65D track & sector
//CPM is only found on 8" disks AFAIK
TCHAR * GetCPMDirectory(unsigned char *pMem, DISKTYPE dt, OSIDirMap* pDirMap)
{
	if (!pMem || (dt != OS65D8 && dt != OS65D5 && dt != DOS658 && dt != DOS655 ))
		return NULL;			

	TCHAR * pBuf = NULL;
	unsigned char * pTrack = pMem;


	OSISector sect;
	memset(&sect, 0, sizeof(OSISector));

	int nLen = 0;
	int nMaxTrack = 77;
	int nMaxSector = 13;
	int nMaxBlk = 239;	  //max 1K disk block address 74 * 26 * 128 / 1024
	int nInterleave = 6;
	int nTrackRecs = 26; //records per track
	int nBAMlo = DOS65DIRTRACK8; 

	switch (dt)
	{
		case DOS655:
			nInterleave = 1;
			//fall through
		case OS65D5:
			nBAMlo = DOS65DIRTRACK5;
			nLen = TRKSIZ5;
			nMaxTrack = 40;
			nMaxSector = 8;
			nTrackRecs = 16;
			nMaxBlk = 70; 
			break;

		case DOS658:
			nInterleave = 1;
			//fall through
		case OS65D8:
			nLen = TRKSIZ8;
			break;
		default:
			break;
		
	}
	pTrack = pMem + nBAMlo * nLen; //offset to trk containing directory


	if (pDirMap)
	{	//initialize BAM map
		if (dt == OS65D5 || dt == DOS655)
			memset(&pDirMap->BAM128[0], 0xFE, sizeof(pDirMap->BAM128));	  //flag out of bounds values
		memset(&pDirMap->BAM128[0], 0xFF, nLen * nMaxTrack/128 );		  //flag unused blocks
	}


	//This information is based on sample OSI CPM disks & manuals
	//CPM assumes max 64 files per disk	& occupies logical trk 0 record 0-15 (2K)
	//OSI CPM 1.4, 2.2 uses record interleaving of 6. DOS/65 uses no interleaving.
	//To calculate physical record from logical on 8", use logical * 6 mod 26 = physical. If logical > 12, add 1 to physical

	//File Control Block 33 byte (disk FCB = 32 bytes)
	//0 =  entry type (0) not used (or 1-16(+?) user#)	for us E5=deleted else valid (usually 0)
	//1-8  filename	padded with spaces
	//9-11 extension padded with spaces
	//12 Ex  file extent (starts @ 0 for <16K file, one directory entry)
	//13 S1  not used (normally 0)
	//14 S2  file extent high byte (normally 0)	Entry number = ((32*S2)+EX) / (exm+1) where exm is the extent mask value from the Disc Parameter Block
	//15   record count is current size 0 to 128 records of this extent
	//16-31 disk allocation disk block map (1 block = 1K or 8 records of 128 bytes
	//32 next record
	//there may be multiple directory entries with the same name if the file > 16K
	//the extent is incremented for each 16K entry
	//
	// High bits of 9-11 have the following meaning for the extension
	//T1-Read only, T2-system file(hidden), T3=archive


	//A single density disk has a block size of 1K, eack block holds 8 128byte sectors
	//CPM directory occupies the 1st one or more data blocks on the disk.  
	//Since directory on OSI CPM 1.4 is on track 3 the logical track # 0 = physical track # 3 (? probably).

	//Directory entry examples:
	//00 53 54 41 54 20 20 20 20 43 4F 4D 00 00 00 18 03 04 05 00 00 00 00 00 00 00 00 00 00 00 00 00
	//vv  S  T  A  T              C  O  M sq       rc -----------------------------------------------
	// vv->0=valid    rc = # of records occupied         a table that holds the # of each consecutive
	//                     by file                       1K blocks on the disk that holds this file
	//  sq sequence # used if file > 16K 
	// STAT.COM is 24 records long, found on disk blocks 03-05 		goes to trk 4
	// Since each track is 26 sectors (of 128 bytes) each
	// starts blk 3 = 3*8 = 0r24 + trk base = 3r24
	// ends blk 5 = 5*8 = 1r14 + trkbase =  4r14  +8? = 4r22

	//00 50 49 50 20 20 20 20 20 43 4F 4D 00 00 00 38 06 07 08 09 0A 0B 0C 00 00 00 00 00 00 00 00 00 
	//    P  I  P                 C  O  M		   rc
	//  PIP.COM = 56 sectors 7K	disk blocks 6-12	 Goes to trk 4,5,6
	// starts 6*8=48/26 = 1r22 + track base = trk4 rec 22 (offset 0xB00)
	// ends 12*8=96/26 =	3r18 + trk base = track 6, rec 18 (offset 0x0900) + 8? = rec26

	//00 41 53 4D 20 20 20 20 20 43 4F 4D 00 00 00 40 9D 9E 9F A0 A1 A2 A3 A4 00 00 00 00 00 00 00 00 
	//    A  S  M                 C  O  M		   rc
	// ASM.COM 64 sectors, 8K disk blocks 157-164	Goes to trk 51,52,53
	// start block 157 * 8 = 1256 / 26 = 48r8 48*26=1248 1256-1248=8 +trk base 3= TRK 51 record +8 (offest 0x400)	
	// end block 164 * 8 = 1312 / 26 = 50r12 + trk base = 53 rec 12 +8= 53 rec 20 (offset 0xa00) 

	//00 43 38 30 20 20 20 20 20 43 4F 4D 00 00 00 80 02 13 14 15 16 17 1A 1B 20 21 48 49 4A 4B 4C 4D 
	//00 43 38 30 20 20 20 20 20 43 4F 4D 01 00 00 6C 4E 4F 50 51 52 53 54 56 57 58 59 5A 5B 5C 00 00 
	//    C  8  0                 C  O  M sq       rc ------ block map ------------------------------
	//                                             ^- files size is sum of rc's 
	// block 2? trk 0r16 + trkbase = trk 3r16 to r24?
	// block 13 * 8 =104/ 26 = 4r0 + trk base = trk 7r0 

	//trk4-76 = 72 tracks *13*256 bytes = 239616 total size	1872 blks
	int nSectors = OSICountSectors(pTrack, nLen, nBAMlo, sect);

	if (nSectors == 1 && sect.nBlockLen[0] == nMaxSector)
	{
		
		unsigned char dirMem[8192];	//lots'o space
		memset(&dirMem[0], 0, sizeof(dirMem));
		unsigned char *p = &dirMem[0];
		int nRead = 16 * 128;			
		
		//read whole directory deinterleaved  16 records = 2K = 64 directory entries
		if (ReadCPMRecords(pMem, dt, 0, 15, nInterleave, p) > 0)
		{
			int nTotalRecs = 0, nCurrentRecs = 0;
			int nTotalBlks = 0, nCurrentExtents = 0, nCurrentBlks = 0;
			const unsigned char *pName, *pExt;
			TCHAR name[20];
			int direntry = 0;

			pBuf = (TCHAR *) malloc(BUFSIZE);
			if (!pBuf)
				return NULL;
			memset(pBuf, 0, BUFSIZE);
			TCHAR * pCurrent = pBuf;

			if ( dt == DOS655 || dt == DOS658)
				_tcscpy(pCurrent, _T("DOS/65"));
			else
				_tcscpy(pCurrent, _T("OSI CP/M"));
			_tcscat(pCurrent, _T("\r\nRecs Bytes Ex Filename.typ\r\n--------------------------\r\n") );
			pCurrent += _tcslen(pCurrent);
			
			if (p)
			{
				p++; //skip leading null

				while (p <  &dirMem[0] + nRead)
				{
					int recs = *(p+0x0E);
					int ext =  *(p+0x0B);

					//So what happens when there are 2 identical filename with different block maps and no extent?
					//One probaby has a non-zero block lead byte meaning it is not a valid filename

					if (recs > 0 && ext == 0 && *(p-1) == 0 ) //not deleted file, not an extension block, not an invalid dir entry
					{
						int pos = 0;
						memset(name, 0, sizeof(name));
						pName = p;	//save these for extsion name comparisons
						pExt = p+8;
						int i = 0;
						nCurrentRecs = 0;

						//copy name
						for (i = 0; i < 8; i++)
						{
							const unsigned char c = *(p+i);
							if (c > 0x20)  //ignore space padding
							{
								if (pDirMap)
									pDirMap->Name[direntry][pos] = c;
								name[pos++] = c;

							}
							else if (c < 0x20)
							{
								break;
							}

						}
						//and extension
						bool bDot = false;
						
						for (i = 0; i < 3; i++)
						{
							const unsigned char c = *(p+i+8);
							if (c > 0x20)  //ignore space padding
							{
								if (!bDot)
								{
								if (pDirMap)
									pDirMap->Name[direntry][pos] = '.';
									
									name[pos++] = _T('.');
									bDot = true;
								}
								if (pDirMap)
									pDirMap->Name[direntry][pos] = c;
								name[pos++] = c;
							}
							else if (c < 0x20)
							{
								break;
							}
						}
						const unsigned char *e = p;

						//look through all dir entries for full extent
						while (e < &dirMem[0] + nRead )  
						{
							if ( *(e-1) == 0 && !memcmp(e, pName, 8) && !memcmp(e+8, pExt, 3)) //valid dir entries start with 0
							{
								recs = *(e+0x0E);
								ext =  *(e+0x0B);
								nCurrentExtents++;

								int recsleft = recs;
								nTotalRecs += recs;
								nCurrentRecs += recs;
								for (i = 15; i < 32 && recsleft > 0; i++)
								{
									//these are the disk allocation blocks in 1K chunks
									unsigned char n = *(e+i);
									if (n>0 && n <= nMaxBlk)
									{
										nCurrentBlks++;
										nTotalBlks++;
										//convert blk# to record #
										int nRec = n*8;
										if (pDirMap && pDirMap->Start[direntry] <=0)
											pDirMap->Start[direntry] = nRec;
										
										for (int j = 0; j < 8 && j < recsleft && pDirMap; j++)
										{
											int trk = ((nRec+j)/nTrackRecs);  //recs / recs per track = trk
											int nStartRec = trk * nTrackRecs;
											trk += nBAMlo; //set base offset
											pDirMap->End[direntry] = nRec+j;

											
											int off = nRec+j - nStartRec; //remainder = offset to record
											//account for interleave
											int nPos = off * nInterleave % nTrackRecs; 
											if (off > 12 && nInterleave > 1) nPos++;

											int index = (trk * nLen + nPos)/ 128;  //our tracklen != cpm tracklen
											//set all recs used by this block
											//if (pDirMap->BAM128[index] != 0)
											//{
											//	int x = 1;
											//}
											pDirMap->BAM128[index] = (unsigned char) direntry; //map used disk blocks
											
										}									
										if (recsleft >= 8)
											recsleft-=8;
										else
											recsleft = 0;
									}
									if (recsleft == 0)
										break;
								} //for there are disk blocks to process

							}//if name matched
							e = e + 32;	//next entry
						}//while there are entries to read

						name[pos] = 0;

						if (name[0])
						{

#if (_MSC_VER >= 1400) && (defined(UNICODE))
							swprintf(pCurrent, _T("%4d %4dK %2d %s\r\n"), nCurrentRecs, (nCurrentRecs+7)/8, nCurrentExtents,  name );
#elif (defined(UNICODE))							
							_stprintf(pCurrent,  _T("%4d %4dK %2d %s\r\n"), nCurrentRecs, (nCurrentRecs+7)/8, nCurrentExtents,  name );
#else
							sprintf(pCurrent,  _T("%4d %4dK %2d %s\r\n"), nCurrentRecs, (nCurrentRecs+7)/8, nCurrentExtents,  name );

#endif
							
							pCurrent += _tcslen(pCurrent);
							direntry++;
							nCurrentRecs = 0;
							nCurrentExtents = 0;
							nCurrentBlks = 0;
						} 
					
					}
					p = p+ 32;
				}//while there are dir entries
#if (_MSC_VER >= 1400) && (defined(UNICODE))
			swprintf(pCurrent, _T("%d blks of 1872, %dK Free\r\n\r\n"), nTotalRecs, (1872-nTotalRecs+4)/8 );
#else							
			_stprintf(pCurrent, _T("%d blks of 1872, %dK Free\r\n\r\n"), nTotalRecs, (1872-nTotalRecs+4)/8 );
#endif
			pCurrent += _tcslen(pCurrent);

			}

			

		}//if read disk OK

	} //if sector OK

	return pBuf;

}


TCHAR * GetOSIDirectory(unsigned char * pMem, DISKTYPE dt, OSIDirMap* pDirMap)
{
	if (!pMem)
		return NULL;
	
	switch (dt)
	{
		case OS65D5:
		case OS65D8:
			return GetOS65DDir(pMem, dt, pDirMap);
			break;

		case HEXDOS:
			return GetHEXDOSDir(pMem, dt, pDirMap);
			break;

		case DOS655:
		case DOS658:
			return GetDOS65Dir(pMem, dt, pDirMap);
			break;			 
		
		case OS65U8:
			return GetOS65UDir(pMem, dt, pDirMap);
			break;
		default:
			break;
	}

	return NULL;

}

//read p-stream word in lo, hi order
unsigned short ReadPWORD(const unsigned char* &p)
{
	unsigned short  w = (unsigned short )((*p) | (*(p+1)<<8)); 
	p+=2;
	return w;
}
TCHAR * ReadPString(const unsigned char* &p)
{
	static TCHAR buf[256];
	int i=0, nCount = (int) *p;
	p++;
	
	while (--nCount >= 0)
	{
		buf[i++] = *(p++);
	}
	buf[i] = 0;
	return  buf;
}



//GetOS65DFile
//OS65D program file format (Tokenized BASIC, Assembler src, binary executable)
//5 byte preamble 
//00 - Starting address Low
//01 - Starting Address High
//02 - Ending Address Low
//03 - Ending Address High
//04 - Number of Tracks
//<Program Data>@starting address
// Starting Address:  
//   $327E - OS65DV3.2 5.25, OS65DV3.1 5.25 
//   $3A7E - OS65DV3.3 5.25, OS65DV3.2 PlotBasic 5.25 
//	 $317E - OS65DV3.2 8"
//   $3A7E 
// When a BASIC file is loaded by OS65D it is relinked to the current workspace start address 



// OS65D directory
//
// 2 sectors (1 & 2) usually on track 12 (8 for 8") hold the directory information. 
// Each entry requires 8 bytes. There are a total of 64 entries. 
// The entries are formatted as follows:
//  Bytes 0-5 ASCII 6 character filename
//  Byte    6 BCD first track of file
//  Byte    7 BCD Last track of file

// TCHAR * GetOS65DDir(unsigned char * pTrack, DISKTYPE dt)
//
// pTrack is OS65D directory track (12 on 5.258 on 8")
// Return text of directory, memory must be freed by caller
//

TCHAR * GetOS65DDir(unsigned char * pMem, DISKTYPE dt, OSIDirMap* pDirMap)
{
	
	if (!pMem || (dt != OS65D8 && dt != OS65D5))
		return NULL;			

	
	TCHAR * pBuf = NULL;
	unsigned char * pTrack = pMem;

	//dir offset:
	// $43, $57 - track start
	// $12 - track #
	// $58, 
	// $76,- sector start
	// $01, $01 - sector#, sector length
	// Entry 6 bytes = name, trk lo, trk hi (* 32 entries)
	//		(unused entry is "######', 0, 0)
	// $47, $53, 
	// $76,- sector start 
	// $02, $01 2nd sector start & length  (* 32 entries)
	//
	OSISector sect;
	int nLen = 0;
	int nMaxTrack = 0;

	switch (dt)
	{
		case OS65D5:
			pTrack = pMem + DIRTRACK5 * TRKSIZ5;
			nLen = TRKSIZ5;
			nMaxTrack = 40;
			
			break;

		case OS65D8:
			nLen = TRKSIZ8;
			nMaxTrack = 77;//77
			pTrack = pMem +	DIRTRACK8 * TRKSIZ8;
			break;
		default:
			break;
		
	}

		if (pDirMap)
		{	//initialize BAM map
			if (dt == OS65D5)
				memset(&pDirMap->BAM128[0], 0xFE, sizeof(pDirMap->BAM128));	  //flag out of bounds values
			memset(&pDirMap->BAM128[0], 0xFF, nLen * nMaxTrack/128 );		  //flag unused blocks
		}


	//look for PicoDos disk
	//picodos programs vs tracks each track is 8 sectors long
	//1 = 3-6
	//2 = 7-10
	//3 = 11-14
	//4 = 15-18
	//5 = 19-22
	//6 = 23-26
	//7 = 27-30
	//8 = 31-33
	unsigned char szPico[] =  "PICO-DOS " ;
	if (bStrStr( pMem, szPico , 256, sizeof(szPico)-1 ) )
	{	
		//no dir on Pico Dos
		pBuf = (TCHAR *) malloc( 240);
		if (pBuf)
			_tcscpy( pBuf, _T("Pico Dos\r\n--------\r\n"));
		return pBuf;
	}
	
	//look for OSI CPM V1.4 disk  on 8"
	unsigned char szCPM[] =  "OSI CP/M" ;
	//and OSI CPM V2.22
	unsigned char szCPM2[] = "CP/M Vers";
	unsigned char szCPM3[] = "CP/M Rev.";
	unsigned char szDigital[] = "DIGITAL RESEARCH";
	//look for CPM marker in 1st 3 tracks
	if (dt == OS65D8 && (bStrStr( pMem, szCPM , 3 * TRKSIZ8, sizeof(szCPM)-1 ) || 
						 bStrStr( pMem, szCPM2 , 3 * TRKSIZ8, sizeof(szCPM2)-1 ) ||
						 bStrStr( pMem, szCPM3 , 3 * TRKSIZ8, sizeof(szCPM3)-1 ) ||
						 bStrStr( pMem, szDigital , 3 * TRKSIZ8, sizeof(szDigital)-1 )) )
	{

		return GetCPMDirectory(pMem, dt, pDirMap);
	} 
	//look for UCSD p-System disk
	unsigned char szPSYSTEM[] =  "SYSTEM.INTERP" ;
	//look for pSYSTEM marker in 1st  track
	if (bStrStr( pMem, szPSYSTEM , nLen, sizeof(szPSYSTEM)-1 ) )
	{
		//OSI p-System
		//on 5.25 data is stored in OS65D format, as 1 sector per track, each sector is 8 pages long (2K bytes)
		//trk0 = boot, trk1=directory, directory offset starts at 2 blocks (1024 bytes)
		//trk2-39 = data (each block is 512 bytes there are 38*4 (152) blocks per disk 
		//p-System File Types
		//0 = undef
		//1 = xdskfile (not used)
		//2 = codefile - created by compiler, assembler, or linker, contains code segments
		//3 = textfile - all files create be the editor or listings by compiler of linker are this file kind
		//4 = info file
		//5 = data file - any file ceated by a program that is not text will be marked datafile
		//6 = graffile (not used)
		//7 = fotofile (not used)
		//8 = securedir (not used)

		//no sample, but on 8" disk I'd expect 3K/track (6 blocks)
		
		pTrack = pMem +	1 * nLen;
		int nSectors = OSICountSectors(pTrack, nLen, 1, sect);
		
		//Directory is stored in 26 byte blocks, values in lo, hi order
		/* Each directory entry is 26 bytes long. The first directory entry describes the directory itself
		the 1st two blocks are reserved for boot loader

		volume entry (1st entry)
		00 00 06 00 00 00 07 50 41 53 43 41 4C 31 98 00 07 00 00 00 B5 A3 00 00 00 00 
		sssss eeeee xxxxx  ^  P  A  S  C  A  L  1 ttttt fffff xxxxx mmmmm iiiiiiiiiii
		                  len (vol name 1 to 7 chars)       

		ssss=starting blk, eeee=ending blk (1st blk past dir)  tttt=#blks total on disk
		ffff= # files  xxxx,iiii=ignore   
		mmmm= mnt date A3B5 = 1010001110110101  = 1010001 11011 0101  = 81 , 27, 5

		1st dir entry
		06 00 2E 00 05 00 0D 53 59 53 54 45 4D 2E 50 41 53 43 41 4C 4E 50 00 02 55 A0
		#1st  #@end ftype  ^  S  Y  S  T  E  M  .  P  A  S  C  A  L  p  p uuuuu ddddd  (1st-end = $28 = 40 blks)
						  13
		A055 = 1010000001010101 = 1010000 00101 0101 = 80, 5, 5
		uuuu=#bytes used in last block (remainder file len)

		2nd dir entry
		2E 00 2F 00 05 00 0F 53 59 53 54 45 4D 2E 4D 49 53 43 49 4E 46 4F C0 00 3C A1 
		#1st  #@end ftype ^   S  Y  S  T  E  M  .  M  I  S  C  I  N  F  O uuuuu ddddd

		A13C = 1010000100111100 = 1010000 10011 1100 = 80, 19, 12

		2F 00 43 00 05 00 0D 53 59 53 54 45 4D 2E 49 4E 54 45 52 50 4E 50 00 02 E5 A0 
							  S  Y  S  T  E  M  .  I  N  T  E  R  P
		A0E5 = 1010000011100101 = 1010000 01110 0101 = 80, 14, 5



		*/
		
		if (nSectors == 1 && sect.nBlockLen[0] >= 8)
		{
			pBuf = (TCHAR *) malloc(BUFSIZE);
			
			if (!pBuf)
				return NULL;
			memset(pBuf, 0, BUFSIZE);
			TCHAR * pCurrent = pBuf;
			const unsigned char * pSector = sect.pData[0];
			const unsigned char * pCur, * pTmp;
			char line[40];
			memset(&line[0], ' ', sizeof(line));
			const TCHAR* Month[16] = { _T("000"), _T("Jan"), _T("Feb"), _T("Mar"), _T("Apr"), _T("May"), _T("Jun"), _T("Jul"), 
									   _T("Aug"), _T("Sep"), _T("Oct"), _T("Nov"), _T("Dec"), _T("DDD"), _T("EEE"), _T("FFF") };
			
			//1st two blocks reserved for boot loader... I find dir starts at track 1, offset 1024
			if (*pSector == 0xE5)
				pSector += 1024;

			int nMaxBlock = 0;
			int nStartBlock = 0;
			int nEndBlock = 0;
			int nFileType = 0;
			int nBlks = 0;
			int nFiles = 0;
			int nDate = 0, yr=0, mn=0,dy=0;
			int nLastBlkUsedBytes = 0;
			int nDiskBlks = 0;
			pCur = pSector;
			
			nStartBlock = ReadPWORD(pCur);	// The first physical block on the disk
			nEndBlock = ReadPWORD(pCur);	// The first block after the directory
			nFileType = ReadPWORD(pCur);    // Ignore, set to zero, corresponding to a UNTYPEDFILE file kind value
			pTmp = pCur;
			TCHAR *szName = ReadPString(pCur);	// The volumes name 1-7 uppercase ASCII chars, length + padding
			pCur = pTmp + 8;				// pad it
			nDiskBlks = ReadPWORD(pCur);	// The physical number of blocks on the volume (152)
			nFiles = ReadPWORD(pCur);		// The number of files on the disk
			pCur += 2;						// Ignore. Set to zero
			nDate = ReadPWORD(pCur);		// The date the volume was last mounted or modified
			//mmmm= mnt date A3B5 = 1010001110110101  = 1010001(yr) 11011(dy) 0101(mn)  = 81 , 27, 5
			mn = nDate & 0x000F;
			nDate >>= 4;
			dy = nDate & 0x1F;
			nDate >>= 5;
			yr = nDate & 0x7F;
			nMaxBlock = nEndBlock; 

			//print volume name
#if (_MSC_VER >= 1400) && (defined(UNICODE))
			swprintf(pCurrent, _T("UCSD p-System\r\n%s:\r\n"), szName );
#else							
			_stprintf(pCurrent, _T("UCSD p-System\r\n%s:\r\n"), szName );
#endif			
			pCurrent += _tcslen(pCurrent); 
			
			//list directory entries
			int j;
			for (j=1; j <= nFiles; j++)
			{
				pCur = pSector + 26*j;
				if (pCur > sect.pData[0] + sect.nBlockLen[0] *256)
					break;
				
				nStartBlock = ReadPWORD(pCur);	//The number of the first block of the file, counting from zero, relative to the start of the disk
				nEndBlock = ReadPWORD(pCur);	//The number of the first block past the end of the file, 
												// counting from zero, relative to the start of the disk
				if (nEndBlock > nMaxBlock)
					nMaxBlock = nEndBlock;
				nFileType = ReadPWORD(pCur);	//The bottom four bits of the 16-bit word describe the file kind.
				pTmp = pCur;
				TCHAR *szName = ReadPString(pCur); //The files name up to 15 characters at least 1 char uppercase ASCII
				pCur = pTmp + 16;				   //padded 

				nLastBlkUsedBytes = ReadPWORD(pCur); //How many bytes of the last block are actually used size=(#blks-1)*512+ lastBlkUsedBytes
				nDate = ReadPWORD(pCur);		//The date the file was last modified
				mn = nDate & 0x000F;
				nDate >>= 4;
				dy = nDate & 0x1F;
				nDate >>= 5;
				yr = nDate & 0x7F;

				nBlks = nEndBlock - nStartBlock;   //this is a fake calculation until blocks are tracked

				//print this entry
#if (_MSC_VER >= 1400) && (defined(UNICODE))
				swprintf(pCurrent, _T("  %-15s  %2d %2d-%s-%2d\r\n"), szName, nBlks, dy, Month[mn], yr );
#else							
				_stprintf(pCurrent, _T("  %-15s  %2d %2d-%s-%2d\r\n"), szName, nBlks, dy, Month[mn], yr );
#endif
				pCurrent += _tcslen(pCurrent);

			}
			//print the disk summary	
#if (_MSC_VER >= 1400) && (defined(UNICODE))
			swprintf(pCurrent, _T(" %d/%d files,  %d unused, %d largest\r\n\r\n"), nFiles, nFiles, nDiskBlks-nMaxBlock, nDiskBlks-nMaxBlock );
#else							
			_stprintf(pCurrent,  _T(" %d/%d files,  %d unused, %d largest\r\n\r\n"), nFiles, nFiles, nDiskBlks-nMaxBlock, nDiskBlks-nMaxBlock);
#endif

			pCurrent += _tcslen(pCurrent);




			return pBuf;
		  
		}
		


	} //UCSD P-SYSTEM


	
	int nSectors = OSICountSectors(pTrack, nLen, -1, sect);
	if (nSectors)
	{
		//14 character filename extension for OS65D
		//uses two file entries, only the 2nd contains track information
		//how can we determine if 14 character filenames are in use?
		// if track ranges are wrong, (ASCII) in BCD track locations

		pBuf = (TCHAR *) malloc(BUFSIZE);
		if (!pBuf)
			return NULL;

		TCHAR * pCurrent = pBuf;

		static TCHAR item[50];
		int nCount = 1;
		
		bool bTrackMap[80] = { 0 };
		bool bTrackMape[80] = {0};
		int trk_collision =0;
		int trke_collision =0; 

		//1st pass decide if valid normal or extended filenames
		int nExtended = 0;
		int nNormal = 0;
		for (int j = 0; j < __min(2, nSectors); j++)
		{
			const unsigned char * pSector = sect.pData[j];
			
			for (int entry = 0; entry < 32; entry++)
			{
				bool bEmpty = true;
				int i;
				bool bASCII = true;

				for ( i = 0; i < 6; i++)
				{
					
					unsigned char c = *(pSector + (8 * entry) + i);
					if (c != 0x23)
						bEmpty = false;
					if (!bEmpty && ( c < 0x20 || c > 0x7F))
						bASCII = false; 
				}


				if (!bEmpty && bASCII)
				{

					unsigned char s1 = *(pSector + 8 * entry + 6);
					unsigned char e1 = *(pSector + 8 * entry + 7);
					int start = BCD2BIN( s1 );
					int end   = BCD2BIN( e1 );
					//s1 & e1 could be ASCII in 14char filename or BCD track range in 6 char filename

					if (!IsValidBCD(s1) || !IsValidBCD(e1) || (end < start) || 
						(end == 0) || (start >= nMaxTrack) || (end >= nMaxTrack) ||
						(s1 >= 0x20 && s1 <= 0x7f && e1 >= 0x20 && e1 <= 0x7f) )
					{
						if ((entry & 1) == 0)
						{
							//if trk range looks like ascii, it may be extended
							if (s1 >= 0x20 && s1 < 0x80 && e1 >= 0x20 && e1 < 0x80)
							{
								unsigned char s2 = *(pSector + 8 * entry + 6+8);
								unsigned char e2 = *(pSector + 8 * entry + 7+8);
								int start2 = BCD2BIN( s2 );
							    int end2   = BCD2BIN( e2 );

								if (IsValidBCD(s2) && IsValidBCD(e2) &&	(end2 >= start2) && (end2 < nMaxTrack)
									&& (end2 != 0) &&  (start2 < nMaxTrack) )
								{
									nExtended++;
									for (int k = start2; k <= end2; k++)
									{
										if (bTrackMape[k])
										{
											trke_collision++;
										}
										else
											bTrackMape[k] = true;
									}

								}

							}
						}

					}
					if (IsValidBCD(s1) && IsValidBCD(e1) && (end >= start) && 
						(end != 0) && (start < nMaxTrack) && (end < nMaxTrack))
					{
						for (int k = start; k <= end; k++)
						{
							if (bTrackMap[k])
							{
								trk_collision++;
							}
							else
								bTrackMap[k] = true;
						}
						
						nNormal++;
					}

				
				}


			}
			


	   }
	   if (nExtended > 1 && nNormal > 1)
	   {
		if (nExtended > (nNormal/2) && trk_collision > trke_collision)
			nNormal = 0;
	   }


		if (nNormal > 0 && nNormal >= nExtended)
		{
			
			if (trk_collision > 5)
			{
				//invalid disk?	too many track use collisions
				free(pBuf);
				return NULL;

			}			
			
			_tcscpy(pCurrent, _T("OS65D File  Track range\r\n-----------------------\r\n") ); 
			
			for (int j = 0; j < __min(2, nSectors); j++)
			{
				const unsigned char * pSector = sect.pData[j];

				
				for (int entry = 0; entry < 32; entry++)
				{
					bool bEmpty = true;
					int i;

					for (i = 0; i < 40; i++)
						item[i] = _T(' ');
						 
					for ( i = 0; i < 6; i++)
					{
						
						TCHAR c = *(pSector + 8 * entry + i);
						item[i+1] = c;
						
						if (c != _T('#'))
							bEmpty = false;
					}


					if (!bEmpty)
					{
						unsigned char s1 = *(pSector + 8 * entry + 6);
						unsigned char e1 = *(pSector + 8 * entry + 7);

						if (IsValidBCD(s1) && IsValidBCD(e1))
						{
						
							int start = BCD2BIN( *(pSector + 8 * entry + 6) );
							int end   = BCD2BIN( *(pSector + 8 * entry + 7) );

							if ( end != 0 && start <= end )
							{
								nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
								swprintf(&item[13], 10, _T("%02d - %02d\r\n"), start, end);
#else							
								_stprintf(&item[13],  _T("%02d - %02d\r\n"), start, end);
#endif
								_tcscat( pCurrent, item);
								if (pDirMap)
								{
									pDirMap->Start[entry] = start;
									pDirMap->End[entry] = end;
									memcpy(pDirMap->Name[entry], (pSector + 8 * entry), 6);
									pDirMap->Name[entry][6] = 0;

									int idxLo = start * nLen / 128;
									int idxHi = (end * nLen + nLen-1) / 128;
									for ( ; idxLo < idxHi; idxLo++)
									{
										pDirMap->BAM128[idxLo] = (unsigned char) entry;
									} 


								}
							}
						}
					
					}


				}
			}
			if (nCount)
			{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
				swprintf(item, 40, _T("\r\n%02d entries free out of %02d\r\n\r\n"), 65-nCount, 64);	
#else
				_stprintf(item, /*40,*/ _T("\r\n%02d entries free out of %02d\r\n\r\n"), 65-nCount, 64);	
#endif
				_tcscat( pCurrent, item);

			}

		}

		if (nExtended > nNormal )
		{

			if (trke_collision > 5)
			{
				//invalid disk?
				free(pBuf);
				return NULL;
			}

			_tcscpy(pCurrent, _T("OS65De File       Track range\r\n-----------------------------\r\n") ); 
			
			for (int j = 0; j < __min(2, nSectors); j++)
			{
				const unsigned char * pSector = sect.pData[j];

				
				for (int entry = 0; entry < 16; entry++)
				{
					bool bEmpty = true;
					int i;

					for (i = 0; i < 48; i++)
						item[i] = _T(' ');
						 
					for ( i = 0; i < 14; i++)
					{
						
						TCHAR c = *(pSector + 16 * entry + i);
						item[i+1] = c;
						
						if (c != _T('#'))
							bEmpty = false;
					}


					if (!bEmpty)
					{
						unsigned char s1 = *(pSector + 16 * entry + 14);
						unsigned char e1 = *(pSector + 16 * entry + 15);

						if (IsValidBCD(s1) && IsValidBCD(e1))
						{
						
							int start = BCD2BIN( *(pSector + 16 * entry + 14) );
							int end   = BCD2BIN( *(pSector + 16 * entry + 15) );

							if ( end != 0 && start <= end )
							{
								nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
								swprintf(&item[21], 10, _T("%02d - %02d\r\n"), start, end);
#else							
								_stprintf(&item[21],  _T("%02d - %02d\r\n"), start, end);
#endif
								_tcscat( pCurrent, item);
								if (pDirMap)
								{
									pDirMap->Start[entry] = start;
									pDirMap->End[entry] = end;
									memcpy(pDirMap->Name[entry], (pSector + 16 * entry), 14);
									pDirMap->Name[entry][14] = 0;

									int idxLo = start * nLen / 128;
									int idxHi = (end * nLen + nLen-1) / 128;
									for ( ; idxLo < idxHi; idxLo++)
									{
										pDirMap->BAM128[idxLo] = (unsigned char) entry;
									} 

								}

							}
						}
					
					}


				}
			}
			if (nCount)
			{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
				swprintf(item, 50, _T("\r\n%02d entries free out of %02d\r\n\r\n"), 33-nCount, 32);	
#else
				_stprintf(item, /*40,*/ _T("\r\n%02d entries free out of %02d\r\n\r\n"), 33-nCount, 32);	
#endif
				_tcscat( pCurrent, item);

			}


		}
		if (nCount <= 1)
		{
			free(pBuf);
			pBuf = NULL;
		}

	}

	return pBuf;

}

// Decode OS65U directory
//
// OS65U file contents: 1st 16 bytes contains copy of the file directory entry 
// (filename, attibutes, address etc.)
//
// OS65U Directory is stored in file Direc*, there are 16 bytes per file entry
// 6 bytes for filename, 
// 2 bytes password LO, HI,
// 1 byte  file attributes  
// 3 bytes starting pos ll, mm, hh (24 bit address)
// 3 bytes files size   ll, mm, hh (24 bit address)
// 1 bytes semaphore(?) 0x00
//
// Attributes byte has the following meaning
//           bit 7=dir flag
//           bit 6 ?
//           bit 5 ?
//           bit 4 - \
//           bit 3 - -- 3 bits file type: 0 = basic, 1 = data else 'other'
//           bit 2 - /
//           bit 1 \ __ 2 bits access:  0=none 1=Read, 2=Write, 3=R/W
//           bit 0 /
//
// End of directory has null filename 
// A deleted file has 0x01 for filename
// A password is entered as 4 chars uppercase ASCII, 'A' to ']'
// if 1st char is A-P then subtract 'A'(0x41) else = 0x0F  = c1
// if 2nd char is N-] then subtract 'N'(0x4E) else = 0x0F  = c2
// if 3rd char is A-P then subtract 'A' else = 0x0F	 = c3
// if 4th char is N-] then subtract 'N' else = 0x0F	 = c4
// PL = c1 << 4 | c2
// PH = c3 << 4 | c4
// so password 'PASS' = F, F, F, 5 : PL = FF, PH = F5 on directory
// password 'ANAN' = 0, 0, 0, 0
// DIREC* file specifies size of directory in it's length (size) field in pages. 
// A whole track is 0E length,  partial track could be '04' length, leaving space for other files on that disk.
//
// OS65U occupies 1st 25088 (0x6200) bytes or 7 sectors
// The first file is always DIREC*, and since a sector is 3584 bytes and there are 16 bytes per directory entry, 
// a floppy disk can hold the DIRECT* entry and AT MOST 223 additional entries. Smaller sector sizes allow for 
// fewer directory entries.
// --------
//
// TCHAR * GetOS65UDir(unsigned char * pTrack, DISKTYPE dt)
//
// Input: OS65U directory track, and Disk Type (OS65U8)
// Returns contents of directory as text
// Caller is responsible for deleting returned memory
//
TCHAR * GetOS65UDir(unsigned char * pMem, DISKTYPE dt, OSIDirMap* pDirMap)
{

	if (!pMem || dt != OS65U8)
		return NULL;
	
	unsigned char* pTrack = pMem + TRKSIZ8 * DIRTRACK65U;
	
	TCHAR item[128] = {0};
	TCHAR pass[5] = {0};

	int nCount = 0, nPos; 
	int nPad = 0;
	unsigned char key[4] = { 0x80, 0x00, 0x07 };
	unsigned char *pCur = pTrack, *pStart = pTrack;
	unsigned char pwLo, pwHi;
	unsigned char cAttrib;
	int nStart, nSize;
	unsigned char cSemi;
	const TCHAR * pAccess = _T("");
	const TCHAR * pType = _T("");
	TCHAR *pList = NULL;
	int nNotOS65UDir = 0;
	int nAddrMax = 0;
	int nEntry = 0;
	int nRecover = 0;
	int nTotalEntries = 0;

	if (!pStart)
		return NULL;
	
	if (dt != OS65U8)
		return NULL;

	if (pDirMap)
	{
		memset(&pDirMap->BAM128[0], 0xFF, TRKSIZ8 * (MAXTRK8+1)/128 );
	}

	//determine start of directory data	on track 7
	for (nPad = 0; nPad < 5; nPad++)
	{
		if (!memcmp(pStart+nPad, key, 3) )
			break;
	}
	if (nPad == 5 )
		return NULL;

	pStart = pTrack + nPad+3;
	
	//skip possible $FF and $1F
	for (nPad = 0; nPad < 5; nPad++)
	{
		if (*(pStart+nPad) == 0x1F)
			break;
		if (*(pStart+nPad) == 0x0F)
			break;
		if (*(pStart+nPad) == 0x44)	//1st character is the contents of the files directory entry
		{
			nPad--;
			break;
		}


	}
	if (nPad == 5 ) 
		return NULL;
	
	pList = (TCHAR*) malloc (16384);
	if (!pList)
		return NULL;
		
	pStart += nPad+1;
	if (!strncmp((char *) pStart, "DIREC*", 6))	 //skip file header entry
	{
		pCur = pStart+9;
		nStart = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur += 3;
		nSize = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		nAddrMax = nStart + nSize;
		nTotalEntries = nSize/16-1;
		pStart += 16;
	}

	pCur = pStart;

	_tcscpy(pList,
	// "abcdef Othr2  R/W
	_T("Name    Type  Access Address/Track Length / Sec SecBnd SecLen PassWord\r\n")
	_T("----------------------------------------------------------------------\r\n") );
	
	int direntry = 0;					

	while (*pCur && pCur < pTrack + 0xE00 + nPad)
	{
		nEntry++;
		pCur = pStart + 16*nCount;
		
		//copy 6 byte file name
		for (nPos = 0; nPos < 6; nPos++)
		{
			item[nPos] = *(pCur+nPos);
			if ( (item[nPos] > 0x01 && item[nPos] < 0x20) || (item[nPos] & 0x80) )	 //not really an OS65U name?
				nNotOS65UDir++;
		}
		item[6] = 0;


		pCur += 6;
		pwLo = *pCur++;
		pwHi = *pCur++;

		cAttrib = *pCur++;

		nStart = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur +=3;
		nSize = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur +=3;

		if (nAddrMax < nStart + nSize)
			nAddrMax = nStart + nSize;

		cSemi = *pCur++;
		//determine file access
		switch (cAttrib & 0x03)
		{
			case 0:
				pAccess = _T("None ");
				break;
			case 1:
				pAccess = _T("Read ");
				break;
			case 2:
				pAccess = _T("Write");
				break;
			case 3:
				pAccess = _T(" R/W ");
				break;
		}
		//determine file type
		switch ((cAttrib >> 2) & 0x07)
		{
			case 0:
				pType = _T("Data ");
				break;
			case 1:
				pType = _T("Basic");
				break;
			case 2:
				pType = _T("Othr2");
				break;
			case 3:
				pType = _T("Othr3");
				break;
			case 4:
				pType = _T("Othr4");
				break;
			case 5:
				pType = _T("Othr5");
				break;
			case 6:
				pType = _T("Othr6");
				break;
			case 7:
				pType = _T("Othr7");
				break;
		}

		//high bit is directory flag
		if (cAttrib & 0x80)
			pType = _T("Direc");

		//decode file password.... shhh this is (supposedly) a secret
		if (pwLo == 0xFF && pwHi == 0xF5)
			_tcscpy(pass, _T("PASS"));
		else
		{
			pass[0] = (TCHAR) (_T('A') + (pwLo>>4));
			pass[1] = (TCHAR) (_T('N') + (pwLo & 0x0F));
			pass[2] = (TCHAR) (_T('A') + (pwHi>>4));
			pass[3] = (TCHAR) (_T('N') + (pwHi & 0x0F));
			pass[4] = 0;
		}
		bool bSec = nStart/3584*3584 == nStart;
		bool bLen = nSize/3584*3584 == nSize;
	    

		
		//item 
		if (*item == 0x01)
		{
			_tcscpy((TCHAR*) item, _T("(del) "));
			nRecover += nSize;
		}
		else
		{

			if (pDirMap)
			{
				int i;
				for (i=0; i < 7; i++)
					pDirMap->Name[direntry][i]= (unsigned char) item[i];
				
				//OS65U trk = 3584, there are 28 128byte blocks
				
				for (i = nStart/128; i < (nStart+nSize)/128; i++)
				{
					int trk = i/28;
					int off = i - trk*28;

					int idx = trk * TRKSIZ8 / 128 + off;
					pDirMap->BAM128[idx] = (unsigned char) direntry;
				}
			}
			direntry++;
		}
		nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
		swprintf( ((TCHAR *) &item)+6, 120, _T("  %s  %s %6d / %02d   %6d / %02d  %s    %s    %s\r\n"), pType,pAccess, nStart, nStart/3584, nSize, (nSize+3583)/3584,
		bSec ? _T("yes") : _T("no "), bLen ? _T("yes") : _T("no "), &pass[0] );
#else
		_stprintf( ((TCHAR *) &item)+6, /*120,*/ _T("  %s  %s %6d / %02d   %6d / %02d   %s    %s    %s\r\n"), pType,pAccess, nStart, nStart/3584, nSize, (nSize+3583)/3584,
		bSec ? _T("yes") : _T("no "), bLen ? _T("yes") : _T("no "), &pass[0] );
#endif
		_tcscat( pList, (TCHAR *)item);

	}
	
	_tcscat( pList, _T("\r\n"));
	if (nRecover)
	{
		_stprintf( item, _T(" %d Bytes Recoverable\r\n"), nRecover );
		_tcscat( pList, (TCHAR *)item);
	}

	if (nAddrMax <= 275968) //max 8" 77 trk floppy storage
	{
		int nFree = (275968 - nAddrMax);
		if (nFree >= 0)
		{
			_stprintf( item, _T(" %d Bytes Free\r\n %d Sector(s) Free\r\n"),nFree, nFree/3584);
			_tcscat( pList, (TCHAR *)item);

		}
	}
	if (nTotalEntries)
	{
			_stprintf( item, _T(" %d Files Defined of %d Possible\r\n\r\n"), nEntry, nTotalEntries);
			_tcscat( pList, (TCHAR *)item);
	
	}
	if (nNotOS65UDir > 10)
	{
		//too many invalid characters ... this can't be a real directory
		free(pList);
		pList = NULL;
	}



	return pList;
}
/* HEXDOS dir loads to $0B01
Stored as linked BASIC listing, line number is track number
Filenames can be variable length
A machine language file starts with $, the 1st two bytes are load & start address, lo byte 1st

        $0B01
         v
57 01    0C 0B 02 00 48 45 58 41 53 4D 00 12 0B 09 00 A5 00 00 00
TRK 01   ----- ~~~~~ =================    ----- NxtTrk *    -----
            |  Trk#   H  E  X  A  S  M     ^  |              ^
            |                              |  |              |
            --------------------------------  ----------------
Since the directory is stored as a BASIC file, Keywords are tokenized in directory listings
so any characters with high bit set are tokenized keywords.
* denotes 1st free track

*/
//
// TCHAR * GetHEXDOSDir(unsigned char * pTrack, DISKTYPE dt)
//
// Input: Hexdos directory track, and Disk Type 
// Returns contents of directory as text
// Caller is responsible for deleting returned memory
//
TCHAR * GetHEXDOSDir(unsigned char * pMem, DISKTYPE dt, OSIDirMap* pDirMap)
{

	if (!pMem || dt != HEXDOS)
		return NULL;

	unsigned char* pTrack = pMem + TRKSIZ5;


	// OSI BASIC in ROM	token list
	const TCHAR *keywordROM[] = 
	{
		_T("END"),  _T("FOR"),  _T("NEXT"), _T("DATA"),   _T("INPUT"),  _T("DIM"),    _T("READ"),  _T("LET"),
		_T("GOTO"), _T("RUN"),  _T("IF"),   _T("RESTORE"),_T("GOSUB"),  _T("RETURN"), _T("REM"),   _T("STOP"),
		_T("ON"),   _T("NULL"), _T("WAIT"), _T("LOAD"),   _T("SAVE"),   _T("DEF"),    _T("POKE"),  _T("PRINT"),
		_T("CONT"), _T("LIST"), _T("CLEAR"),_T("NEW"),    _T("TAB("),   _T("TO"),     _T("FN"),    _T("SPC("),   
		_T("THEN"), _T("NOT"),  _T("STEP"), _T("+"),      _T("-"),      _T("*"),      _T("/"),     _T("^"),  
		_T("AND"),  _T("OR"),   _T(">"),    _T("="),      _T("<"),      _T("SGN"),    _T("INT"),   _T("ABS"),
		_T("USR"),  _T("FRE"),  _T("POS"),	_T("SQR"),    _T("RND"),    _T("LOG"),    _T("EXP"),   _T("COS"),
		_T("SIN"),  _T("TAN"),  _T("ATN"),  _T("PEEK"),   _T("LEN"),    _T("STR$"),   _T("VAL"),   _T("ASC"),
		_T("CHR$"), _T("LEFT$"),_T("RIGHT$"),_T("MID$")

	};

	TCHAR * pList = NULL;
	int nEntry = 0;
	if (pTrack)
	{
		//dir offset:
		// $57, $01 - track start
		// <2K sector data>
		int nPos = OSISeekTrackStart(pTrack, TRKSIZ5);

	    if (nPos >= 0)
		{
			int base = 0x0B01;
			if (pTrack[nPos] != 0x57 && pTrack[nPos+1] != 0x01)
				return NULL;
			nPos+=2;
			unsigned char* p = pTrack+nPos;

			int offset = 0;
			TCHAR szLine[1024];

			pList = (TCHAR*) malloc (16384);
			if (!pList)
				return NULL;

			_tcscpy(pList,
			_T("HEXDOS Track, Filename\r\n")
			_T("----------------------\r\n") );
 
			int nLineStart = 0;
			while (pList)
			{
				*szLine = 0;
				offset = *p + (*(p+1) << 8);
				if (offset == 0)
					break;
				int nTrack = *(p+2) + (*(p+3) << 8);
				if (nTrack < 40)
				{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
					swprintf( szLine, _T("%02d, "), nTrack );
#else
					_stprintf( szLine, _T("%02d, "), nTrack );
#endif				
					nLineStart = _tcslen(szLine);
					//decode BASIC listing
					unsigned char * pText = p+4;
					while (*pText)
					{
						if ( (*pText) & 0x80)
						{
							//keyword
							int i = *pText & 0x7F;
							if (i <	sizeof(keywordROM)/sizeof(keywordROM[0]))
							{
								_tcscat(szLine, keywordROM[i]);
							}
						}
						else
						{
							TCHAR buf[2] = {0};
							buf[0] = *pText; 
							_tcscat(szLine, buf);	
						}
						pText++;
					}
					if (pDirMap)
					{
						bool bEndTrack = false;
						int i;
						for (i = nLineStart; i < sizeof(pDirMap->Name[1]); i++)
						{
							unsigned char c = (unsigned char ) szLine[i];
							pDirMap->Name[nEntry][i] = c;
							if (c == '*')
								bEndTrack = true;
							if (!szLine[i])
								break;
							
						}
						if (!bEndTrack)
						{
							pDirMap->Start[nEntry] = nTrack;
							pDirMap->End[nEntry] = nTrack; //assume 1 track long for now
							if (nEntry > 0)
							{
								pDirMap->End[nEntry-1]= nTrack-1; //fix up previous end track
							}
							nEntry++;
						}
					}

					_tcscat(szLine, _T("\r\n"));

				}
				_tcscat(pList, szLine);

				if (offset - base > 2048 || offset < base)
				{
					break;	
				} 
				//point to next basic line
				unsigned char* pNext = pTrack+nPos+ (offset - base);
				if ( pNext > p)
					p = pNext;
				else
					break; //reverse traverse?
			}
			_tcscat(pList, _T("\r\n"));
		}

		
	}
	return pList;

}

//Dos 65 uses same directory layout and track math as CP/M, however it is non-interleaved
//and can be forund on 5.25" and 8" floppies
	    	  
TCHAR * GetDOS65Dir(unsigned char * pDisk, DISKTYPE dt, OSIDirMap* pDirMap)
{
	return GetCPMDirectory(pDisk, dt, pDirMap);
	
}